/* 
 * tclIOCmd.c --
 *
 *	Contains the definitions of all Tcl commands relating to IO.
 *
 * Copyright (c) 1995-1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

static char sccsid[] = "@(#) tclIOCmd.c 1.65 96/01/30 12:40:27";

#include	"tclInt.h"	/* Internal definitions for Tcl. */
#include	"tclPort.h"	/* Portability features for Tcl. */

/*
 * Return at most this number of bytes in one call to Tcl_ReadChannel:
 */

#define	TCL_READ_CHUNK_SIZE	4096

/*
 * Identifiers for the different protocols understood by Tcl_SockCmd.
 */

#define	TCL_SOCKPROTO_TCP	0
#define TCL_SOCKPROTO_UDP	1
#define TCL_SOCKPROTO_BCAST	2

/*
 * Callback structure for accept callback in a TCP server.
 */

typedef struct AcceptCallback {
    char *script;			/* Script to invoke. */
    Tcl_Interp *interp;			/* Interpreter in which to run it. */
} AcceptCallback;

/*
 * Static functions for this file:
 */

static void	AcceptCallbackProc _ANSI_ARGS_((ClientData callbackData,
	            Tcl_Channel chan, char *address));
static void	RegisterTcpServerInterpCleanup _ANSI_ARGS_((Tcl_Interp *interp,
	            AcceptCallback *acceptCallbackPtr));
static void	TcpAcceptCallbacksDeleteProc _ANSI_ARGS_((
		    ClientData clientData, Tcl_Interp *interp));
static void	TcpServerCloseCallback _ANSI_ARGS_((ClientData callbackData));
static void	UnregisterTcpServerInterpCleanup _ANSI_ARGS_((
		    Tcl_Interp *interp, AcceptCallback *acceptCallbackPtr));

/*
 *----------------------------------------------------------------------
 *
 * TcpAcceptCallbacksDeleteProc --
 *
 *	Assocdata cleanup routine called when an interpreter is being
 *	deleted to set the interp field of all the accept callback records
 *	registered with	the interpreter to NULL. This will prevent the
 *	interpreter from being used in the future to eval accept scripts.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Deallocates memory and sets the interp field of all the accept
 *	callback records to NULL to prevent this interpreter from being
 *	used subsequently to eval accept scripts.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
TcpAcceptCallbacksDeleteProc(clientData, interp)
    ClientData clientData;	/* Data which was passed when the assocdata
                                 * was registered. */
    Tcl_Interp *interp;		/* Interpreter being deleted - not used. */
{
    Tcl_HashTable *hTblPtr;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch hSearch;
    AcceptCallback *acceptCallbackPtr;

    hTblPtr = (Tcl_HashTable *) clientData;
    for (hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch);
             hPtr != (Tcl_HashEntry *) NULL;
             hPtr = Tcl_FirstHashEntry(hTblPtr, &hSearch)) {
        acceptCallbackPtr = (AcceptCallback *) Tcl_GetHashValue(hPtr);
        acceptCallbackPtr->interp = (Tcl_Interp *) NULL;
        Tcl_DeleteHashEntry(hPtr);
    }
    Tcl_DeleteHashTable(hTblPtr);
    ckfree((char *) hTblPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * RegisterTcpServerInterpCleanup --
 *
 *	Registers an accept callback record to have its interp
 *	field set to NULL when the interpreter is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When, in the future, the interpreter is deleted, the interp
 *	field of the accept callback data structure will be set to
 *	NULL. This will prevent attempts to eval the accept script
 *	in a deleted interpreter.
 *
 *----------------------------------------------------------------------
 */

static void
RegisterTcpServerInterpCleanup(interp, acceptCallbackPtr)
    Tcl_Interp *interp;		/* Interpreter for which we want to be
                                 * informed of deletion. */
    AcceptCallback *acceptCallbackPtr;
    				/* The accept callback record whose
                                 * interp field we want set to NULL when
                                 * the interpreter is deleted. */
{
    Tcl_HashTable *hTblPtr;	/* Hash table for accept callback
                                 * records to smash when the interpreter
                                 * will be deleted. */
    Tcl_HashEntry *hPtr;	/* Entry for this record. */
    int new;			/* Is the entry new? */

    hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp,
            "tclTCPAcceptCallbacks",
            NULL);
    if (hTblPtr == (Tcl_HashTable *) NULL) {
        hTblPtr = (Tcl_HashTable *) ckalloc((unsigned) sizeof(Tcl_HashTable));
        Tcl_InitHashTable(hTblPtr, TCL_ONE_WORD_KEYS);
        (void) Tcl_SetAssocData(interp, "tclTCPAcceptCallbacks",
                TcpAcceptCallbacksDeleteProc, (ClientData) hTblPtr);
    }
    hPtr = Tcl_CreateHashEntry(hTblPtr, (char *) acceptCallbackPtr, &new);
    if (!new) {
        panic("RegisterTcpServerCleanup: damaged accept record table");
    }
    Tcl_SetHashValue(hPtr, (ClientData) acceptCallbackPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * UnregisterTcpServerInterpCleanup --
 *
 *	Unregister a previously registered accept callback record. The
 *	interp field of this record will no longer be set to NULL in
 *	the future when the interpreter is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Prevents the interp field of the accept callback record from
 *	being set to NULL in the future when the interpreter is deleted.
 *
 *----------------------------------------------------------------------
 */

static void
UnregisterTcpServerInterpCleanup(interp, acceptCallbackPtr)
    Tcl_Interp *interp;		/* Interpreter in which the accept callback
                                 * record was registered. */
    AcceptCallback *acceptCallbackPtr;
    				/* The record for which to delete the
                                 * registration. */
{
    Tcl_HashTable *hTblPtr;
    Tcl_HashEntry *hPtr;

    hTblPtr = (Tcl_HashTable *) Tcl_GetAssocData(interp,
            "tclTCPAcceptCallbacks", NULL);
    if (hTblPtr == (Tcl_HashTable *) NULL) {
        return;
    }
    hPtr = Tcl_FindHashEntry(hTblPtr, (char *) acceptCallbackPtr);
    if (hPtr == (Tcl_HashEntry *) NULL) {
        return;
    }
    Tcl_DeleteHashEntry(hPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_PutsCmd --
 *
 *	This function implements the "puts" Tcl command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Produces output on a channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_PutsCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to puts on. */
    int i;				/* Counter. */
    int newline;			/* Add a newline at end? */
    char *channelId;			/* Name of channel for puts. */
    int result;				/* Result of puts operation. */
    int mode;				/* Mode in which channel is opened. */

    i = 1;
    newline = 1;
    if ((argc >= 2) && (strcmp(argv[1], "-nonewline") == 0)) {
	newline = 0;
	i++;
    }
    if ((i < (argc-3)) || (i >= argc)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" ?-nonewline? ?channelId? string\"", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * The code below provides backwards compatibility with an old
     * form of the command that is no longer recommended or documented.
     */

    if (i == (argc-3)) {
	if (strncmp(argv[i+2], "nonewline", strlen(argv[i+2])) != 0) {
	    Tcl_AppendResult(interp, "bad argument \"", argv[i+2],
		    "\": should be \"nonewline\"", (char *) NULL);
	    return TCL_ERROR;
	}
	newline = 0;
    }
    if (i == (argc-1)) {
	channelId = "stdout";
    } else {
	channelId = argv[i];
	i++;
    }
    channelId = TclSubstChannelName(channelId);
    chan = Tcl_GetChannel(interp, channelId, &mode);
    if (chan == (Tcl_Channel) NULL) {
        return TCL_ERROR;
    }
    if ((mode & TCL_WRITABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", channelId,
                "\" wasn't opened for writing", (char *) NULL);
        return TCL_ERROR;
    }
    
    result = Tcl_Write(chan, argv[i], -1);
    if (result < 0) {
        Tcl_AppendResult(interp, "puts: ", Tcl_GetChannelName(chan), ": ",
                Tcl_PosixError(interp), (char *) NULL);
        return TCL_ERROR;
    }
    if (newline != 0) {
        result = Tcl_Write(chan, "\n", 1);
        if (result < 0) {
            Tcl_AppendResult(interp, "puts: ", Tcl_GetChannelName(chan), ": ",
                    Tcl_PosixError(interp), (char *) NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_FlushCmd --
 *
 *	Implements the Tcl "flush" command. May flush output queued for
 *	a channel.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May cause output to appear on the specified channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_FlushCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to flush on. */
    char *channelId;			/* Name of channel to flush. */
    int result;				/* Result of call to channel
                                         * level function. */
    int mode;				/* Mode in which channel is opened. */

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" channelId\"", (char *) NULL);
	return TCL_ERROR;
    }
    channelId = TclSubstChannelName(argv[1]);
    chan = Tcl_GetChannel(interp, channelId, &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_WRITABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for writing", (char *) NULL);
        return TCL_ERROR;
    }
    
    result = Tcl_Flush(chan);
    if (result != TCL_OK) {
        Tcl_AppendResult(interp, "flush: ", Tcl_GetChannelName(chan), ": ",
                Tcl_PosixError(interp), (char *) NULL);
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetsCmd --
 *
 *	Implements the Tcl "gets" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May cause output to be flushed. May consume input from channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_GetsCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to read from. */
    char *channelId;			/* Name of channel to read from. */
    char *varName;			/* Assign to this variable? */
    char buf[128];			/* Buffer to store string
                                         * representation of how long
                                         * a line was read. */
    int lineLen;			/* To hold line length read. */
    Tcl_DString ds;			/* Dynamic string to hold the
                                         * buffer for the line just read. */
    char *linePtr;			/* Points at buffer in dstring. */
    int mode;				/* Mode in which channel is opened. */

    if ((argc != 2) && (argc != 3)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" channelId ?varName?\"", (char *) NULL);
	return TCL_ERROR;
    }
    channelId = TclSubstChannelName(argv[1]);
    chan = Tcl_GetChannel(interp, channelId, &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
    
    if (argc != 3) {
        varName = (char *) NULL;
    } else {
        varName = argv[2];
    }
    Tcl_DStringInit(&ds);
    lineLen = Tcl_Gets(chan, &ds);
    if (lineLen < 0) {
        Tcl_DStringFree(&ds);
        if (Tcl_Eof(chan) || Tcl_InputBlocked(chan)) {
            if (varName != (char *) NULL) {
                Tcl_AppendResult(interp, "-1", (char *) NULL);
            }
            return TCL_OK;
        }
        Tcl_AppendResult(interp, "gets: ", Tcl_GetChannelName(chan), ": ",
                Tcl_PosixError(interp), (char *) NULL);
        return TCL_ERROR;
    }

    linePtr = Tcl_DStringValue(&ds);
    if (varName == (char *) NULL) {
        Tcl_AppendResult(interp, linePtr, (char *) NULL);
    } else {
        if (Tcl_SetVar(interp, varName, linePtr, TCL_LEAVE_ERR_MSG) ==
                NULL) {
            Tcl_DStringFree(&ds);
            return TCL_ERROR;
        }
        Tcl_ResetResult(interp);
        sprintf(buf, "%d", lineLen);
        Tcl_AppendResult(interp, buf, (char *) NULL);
    }
    Tcl_DStringFree(&ds);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_ReadCmd --
 *
 *	Implements the Tcl "read" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May cause output to be flushed. May consume input from channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_ReadCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to read from. */
    char *channelId;			/* Name of channel to read from. */
    int newline, i;			/* Discard newline at end? */
    int toRead;				/* How many bytes to read? */
    char *bufPtr;			/* Temporary buffer used to
                                         * store chunks read. */
    int toReadNow;			/* How many bytes to attempt to
                                         * read in the current iteration? */
    int charactersRead;			/* How many characters were read? */
    int charactersReadNow;		/* How many characters were read
                                         * in this iteration? */
    int mode;				/* Mode in which channel is opened. */
    int shortRead;			/* When we get a short read this is
                                         * set to non-zero to cause the reading
                                         * loop to stop. */

    if ((argc != 2) && (argc != 3)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" channelId ?numBytes?\" or \"", argv[0],
		" ?-nonewline? channelId\"", (char *) NULL);
	return TCL_ERROR;
    }
    i = 1;
    newline = 0;
    if (strcmp(argv[i], "-nonewline") == 0) {
	newline = 1;
	i++;
    }

    if (i == argc) {
        Tcl_AppendResult(interp, "read: missing channel name", (char *)
                NULL);
        return TCL_ERROR;
    }

    channelId = TclSubstChannelName(argv[i]);
    chan = Tcl_GetChannel(interp, channelId, &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[i],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
    
    i++;	/* Consumed channel name. */

    /*
     * Compute how many bytes to read, and see whether the final
     * newline should be dropped.
     */

    if (i == argc) {
	toRead = INT_MAX;
    } else if ((newline == 0) && (isdigit((unsigned char) (argv[i][0])))) {
	if (Tcl_GetInt(interp, argv[i], &toRead) != TCL_OK) {
	    return TCL_ERROR;
	}
    } else {
	toRead = INT_MAX;

	/*
	 * The code below provides backward compatibility for an
	 * archaic earlier version of this command.
	 */

        if (strcmp(argv[i], "nonewline") == 0) {
            newline = 1;
        } else {
            Tcl_AppendResult(interp, "bad argument \"", argv[i],
                    "\": should be \"nonewline\"", (char *) NULL);
            return TCL_ERROR;
        }
    }

    bufPtr = ckalloc((unsigned) TCL_READ_CHUNK_SIZE + 1);
    
    for (charactersRead = 0, shortRead = 0;
             (charactersRead < toRead) && (!shortRead);
             charactersRead += charactersReadNow) {
        toReadNow = toRead - charactersRead;
        if (toReadNow > TCL_READ_CHUNK_SIZE) {
            toReadNow = TCL_READ_CHUNK_SIZE;
        }
        charactersReadNow = Tcl_Read(chan, bufPtr, toReadNow);
        if (charactersReadNow < 0) {
            ckfree(bufPtr);

            /*
             * We have to reset the result before storing the error
             * message, because we may have accummulated some input read
             * by previous calls (above) to Tcl_Read.
             */
            
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "read: ", Tcl_GetChannelName(chan), ": ",
                    Tcl_PosixError(interp), (char *) NULL);
            return TCL_ERROR;
        }

        /*
         * Ensure that the buffer ends with a NULL so that it can safely
         * be given to Tcl_AppendResult.
         */

        bufPtr[charactersReadNow] = '\0';
        Tcl_AppendResult(interp, bufPtr, (char *) NULL);

        /*
         * If we had a short read it means that we have either EOF
         * or BLOCKED on the channel, so signal the loop to stop
         * reading.
         */

        if (charactersReadNow < toReadNow) {
            shortRead = 1;
        }
    }
    ckfree(bufPtr);

    /*
     * If there is no input string it means that we got EOF or BLOCKED
     * right at the first character that was read from the channel.
     */
    
    if (charactersRead == 0) {
        Tcl_ResetResult(interp);
        return TCL_OK;
    }

    /*
     * If requested, remove the last newline in the channel if at EOF.
     */
    
    if ((newline) && (Tcl_Eof(chan)) &&
            (interp->result[charactersRead-1] == '\n')) {
        interp->result[charactersRead-1] = '\0';
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclUnsupported0Cmd --
 *
 *	Implements the "unsupported0" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May copy a chunk from one channel to another.
 *
 *----------------------------------------------------------------------
 */

int
TclUnsupported0Cmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Interpreter in which both channels
                                         * are defined. */
    int argc;				/* How many arguments? */
    char **argv;			/* The argument strings. */
{
    Tcl_Channel inChan, outChan;
    char *inChanId, *outChanId;
    int requested;
    char *bufPtr;
    int actuallyRead, actuallyWritten, totalRead, toReadNow, mode;
    
    /*
     * Assume we want to copy the entire channel.
     */
    
    requested = INT_MAX;

    if ((argc < 3) || (argc > 4)) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
                argv[0], " inChanId outChanId ?chunkSize?\"", (char *) NULL);
        return TCL_ERROR;
    }
    inChanId = TclSubstChannelName(argv[1]);
    inChan = Tcl_GetChannel(interp, inChanId, &mode);
    if (inChan == (Tcl_Channel) NULL) {
        return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
    
    outChanId = TclSubstChannelName(argv[2]);
    outChan = Tcl_GetChannel(interp, argv[2], &mode);
    if (outChan == (Tcl_Channel) NULL) {
        return TCL_ERROR;
    }
    if ((mode & TCL_WRITABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[2],
                "\" wasn't opened for writing", (char *) NULL);
        return TCL_ERROR;
    }
    
    if (argc == 4) {
	if (Tcl_GetInt(interp, argv[3], &requested) != TCL_OK) {
	    return TCL_ERROR;
	}
        if (requested < 0) {
            requested = INT_MAX;
        }
    }

    bufPtr = ckalloc((unsigned) TCL_READ_CHUNK_SIZE);
    for (totalRead = 0;
            requested > 0;
            totalRead += actuallyRead, requested -= actuallyRead) {
        toReadNow = requested;
        if (toReadNow > TCL_READ_CHUNK_SIZE) {
            toReadNow = TCL_READ_CHUNK_SIZE;
        }
        actuallyRead = Tcl_Read(inChan, bufPtr, toReadNow);
        if (actuallyRead < 0) {
            ckfree(bufPtr);
	    Tcl_AppendResult(interp, argv[0], ": ", Tcl_GetChannelName(inChan),
		Tcl_PosixError(interp), (char *) NULL);
            return TCL_ERROR;
        }
        if (actuallyRead == 0) {
            ckfree(bufPtr);
            sprintf(interp->result, "%d", totalRead);
            return TCL_OK;
        }
        actuallyWritten = Tcl_Write(outChan, bufPtr, actuallyRead);
        if (actuallyWritten < 0) {
            ckfree(bufPtr);
	    Tcl_AppendResult(interp, argv[0], ": ", Tcl_GetChannelName(outChan),
		Tcl_PosixError(interp), (char *) NULL);
            return TCL_ERROR;
        }
    }
    ckfree(bufPtr);
    
    sprintf(interp->result, "%d", totalRead);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SeekCmd --
 *
 *	Implements the Tcl "seek" command. Moves the current access point
 *	to a new location.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Moves the position of the access point on the specified channel.
 *	May flush queued output.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_SeekCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to tell on. */
    char *channelId;			/* Name of channel to tell on. */
    int offset, mode;			/* Where to seek? */
    int result;				/* Of calling Tcl_Seek. */

    if ((argc != 3) && (argc != 4)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" channelId offset ?origin?\"", (char *) NULL);
	return TCL_ERROR;
    }
    channelId = TclSubstChannelName(argv[1]);
    chan = Tcl_GetChannel(interp, channelId, NULL);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[2], &offset) != TCL_OK) {
	return TCL_ERROR;
    }
    mode = SEEK_SET;
    if (argc == 4) {
	size_t length;
	int c;

	length = strlen(argv[3]);
	c = argv[3][0];
	if ((c == 's') && (strncmp(argv[3], "start", length) == 0)) {
	    mode = SEEK_SET;
	} else if ((c == 'c') && (strncmp(argv[3], "current", length) == 0)) {
	    mode = SEEK_CUR;
	} else if ((c == 'e') && (strncmp(argv[3], "end", length) == 0)) {
	    mode = SEEK_END;
	} else {
	    Tcl_AppendResult(interp, "bad origin \"", argv[3],
		    "\": should be start, current or end", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    result = Tcl_Seek(chan, offset, mode);
    if (result < 0) {
        Tcl_AppendResult(interp, "seek: ", Tcl_GetChannelName(chan), ": ",
                Tcl_PosixError(interp), (char *) NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_TellCmd --
 *
 *	Implements the Tcl "tell" command. Returns the current access
 *	position on the specified channel as a Tcl string.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_TellCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to tell on. */
    char *channelId;			/* The channel name to tell on. */

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" channelId\"", (char *) NULL);
	return TCL_ERROR;
    }
    /*
     * Try to find a channel with the right name and permissions in
     * the IO channel table of this interpreter.
     */
    
    channelId = TclSubstChannelName(argv[1]);
    chan = Tcl_GetChannel(interp, channelId, NULL);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", Tcl_Tell(chan));

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CloseCmd --
 *
 *	Implements the Tcl "close" command. Closes the channel and may
 *	start an asynchronous flush.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May discard queued input; may flush queued output.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_CloseCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to close. */
    char *channelId;			/* Name of channel to close. */

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" channelId\"", (char *) NULL);
	return TCL_ERROR;
    }
    channelId = TclSubstChannelName(argv[1]);
    chan = Tcl_GetChannel(interp, channelId, NULL);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if (Tcl_UnregisterChannel(interp, chan) != TCL_OK) {

        /*
         * This hack checks to see if the result was already set by
         * Tcl_CloseChannel. If so, it does not overwrite that, otherwise
         * it sets its own error result.
         */

        if ((interp->result != (char *) NULL) && (interp->result[0] == '\0')) {
            Tcl_AppendResult(interp, "close: \"<deleted-channel>\": ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_FModeCmd --
 *
 *	Implements the Tcl "fconfigure" command. This sets modes on an IO
 *	channel.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May modify the behavior of an IO channel
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_FModeCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to set a mode on. */
    char *channelId;			/* Name of channel to set mode on. */
    int result;				/* Of Tcl_Set/GetChannelOption. */
    char *optionValue;			/* Value retrieved by a call to
                                         * Tcl_GetChannelOption. */
    int i;				/* Iterate over arg-value pairs. */

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                " channelId ?optionName? ?value? ?optionName value?...\"",
                (char *) NULL);
        return TCL_ERROR;
    }
    channelId = TclSubstChannelName(argv[1]);
    chan = Tcl_GetChannel(interp, channelId, NULL);
    if (chan == (Tcl_Channel) NULL) {
        return TCL_ERROR;
    }
    if (argc == 2) {
        Tcl_AppendElement(interp, "-blocking");
        Tcl_AppendElement(interp, Tcl_GetChannelOption(chan, "-blocking"));

        Tcl_AppendElement(interp, "-linemode");
        Tcl_AppendElement(interp, Tcl_GetChannelOption(chan, "-linemode"));

        Tcl_AppendElement(interp, "-translation");
        Tcl_AppendElement(interp, Tcl_GetChannelOption(chan, "-translation"));

        return TCL_OK;
    }
    if (argc == 3) {
        optionValue = Tcl_GetChannelOption(chan, argv[2]);
        if (optionValue == NULL) {
            Tcl_AppendResult(interp, "fconfigure: ", Tcl_GetChannelName(chan),
                    ": ", Tcl_PosixError(interp), (char *) NULL);
            return TCL_ERROR;
        }
        sprintf(interp->result, "%s", optionValue);
        return TCL_OK;
    }
    if ((argc % 2) == 1) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                " channelId ?optionName? ?value? ?optionName value?...\"",
                (char *) NULL);
        return TCL_ERROR;
    }
    for (i = 3; i < argc; i += 2) {
        result = Tcl_SetChannelOption(interp, chan, argv[i-1], argv[i]);
        if (result != TCL_OK) {
            return result;
        }
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_EofCmd --
 *
 *	Implements the Tcl "eof" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Sets interp->result to "0" or "1" depending on whether the
 *	specified channel has an EOF condition.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_EofCmd(unused, interp, argc, argv)
    ClientData unused;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to query for EOF. */
    char *channelId;			/* Name of channel to query. */
    int mode;				/* Mode in which channel is opened. */

    if (argc != 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                " channelId\"", (char *) NULL);
        return TCL_ERROR;
    }
    channelId = TclSubstChannelName(argv[1]);
    chan = Tcl_GetChannel(interp, channelId, &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", Tcl_Eof(chan) ? 1 : 0);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_FBlockedCmd --
 *
 *	Implements the Tcl "fblocked" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Sets interp->result to "0" or "1" depending on whether the
 *	a preceding input operation on the channel would have blocked.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_FBlockedCmd(unused, interp, argc, argv)
    ClientData unused;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to query for blocked. */
    char *channelId;			/* Name of channel to query. */
    int mode;				/* Mode in which channel was opened. */

    if (argc != 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                " channelId\"", (char *) NULL);
        return TCL_ERROR;
    }
    channelId = TclSubstChannelName(argv[1]);
    chan = Tcl_GetChannel(interp, channelId, &mode);
    if (chan == (Tcl_Channel) NULL) {
        return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
        
    sprintf(interp->result, "%d", Tcl_InputBlocked(chan) ? 1 : 0);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenCmd --
 *
 *	This procedure is invoked to process the "open" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_OpenCmd(notUsed, interp, argc, argv)
    ClientData notUsed;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    int pipeline, prot;
    char *mode;
    int cmdArgc;
    char **cmdArgv;
    Tcl_Channel chan;

    if ((argc < 2) || (argc > 4)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" fileName ?access? ?permissions?\"", (char *) NULL);
	return TCL_ERROR;
    }
    prot = 0666;
    if (argc == 2) {
	mode = "r";
    } else {
	mode = argv[2];
	if (argc == 4) {
	    if (Tcl_GetInt(interp, argv[3], &prot) != TCL_OK) {
		return TCL_ERROR;
	    }
	}
    }

    pipeline = 0;
    if (argv[1][0] == '|') {
	pipeline = 1;
    }

    /*
     * Open the file or create a process pipeline.
     */

    if (!pipeline) {
        chan = Tcl_OpenFileChannel(interp, argv[1], mode, prot);
        if (chan == (Tcl_Channel) NULL) {
            return TCL_ERROR;
        }
    } else {
        if (Tcl_SplitList(interp, argv[1]+1, &cmdArgc, &cmdArgv) != TCL_OK) {
            return TCL_ERROR;
        }
        chan = Tcl_OpenCommandChannel(interp, cmdArgc, cmdArgv, mode);
        ckfree((char *) cmdArgv);
        if (chan == (Tcl_Channel) NULL) {
            return TCL_ERROR;
        }
    }
    Tcl_RegisterChannel(interp, chan);
    Tcl_AppendResult(interp, Tcl_GetChannelName(chan), (char *) NULL);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * AcceptCallbackProc --
 *
 *	This callback is invoked by the TCP channel driver when it
 *	accepts a new connection from a client on a server socket.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Whatever the script does.
 *
 *----------------------------------------------------------------------
 */

static void
AcceptCallbackProc(callbackData, chan, address)
    ClientData callbackData;		/* The data stored when the callback
                                         * was created in the call to
                                         * Tcl_OpenTcpServer. */
    Tcl_Channel chan;			/* Channel for the newly accepted
                                         * connection. */
    char *address;			/* Address of client that was
                                         * accepted, in four-byte network
                                         * address format. */    
{
    AcceptCallback *acceptCallbackPtr;
    Tcl_Interp *interp;
    char *script;
    int result;

    acceptCallbackPtr = (AcceptCallback *) callbackData;

    /*
     * Check if the callback is still valid; the interpreter may have gone
     * away, this is signalled by setting the interp field of the callback
     * data to NULL.
     */
    
    if (acceptCallbackPtr->interp != (Tcl_Interp *) NULL) {

        script = acceptCallbackPtr->script;
        interp = acceptCallbackPtr->interp;
        
        Tcl_Preserve((ClientData) script);

        Tcl_RegisterChannel(interp, chan);
        result = Tcl_VarEval(interp, script, " ", Tcl_GetChannelName(chan),
                " ", address, (char *) NULL);
        if (result != TCL_OK) {
            Tcl_BackgroundError(interp);
        }

        Tcl_Release((ClientData) script);
    } else {

        /*
         * The interpreter is gone, so there is no useful way to utilize
         * the client socket - just close it.
         */

        Tcl_Close(chan);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TcpServerCloseCallback --
 *
 *	This callback is called when the TCP server channel for which it
 *	was registered is being closed. It informs the interpreter in
 *	which the accept script is evaluated (if that interpreter still
 *	exists) that this channel no longer needs to be informed if the
 *	interpreter is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	In the future, if the interpreter is deleted this channel will
 *	no longer be informed.
 *
 *----------------------------------------------------------------------
 */

static void
TcpServerCloseCallback(callbackData)
    ClientData callbackData;	/* The data passed in the call to
                                 * Tcl_CreateCloseHandler. */
{
    AcceptCallback *acceptCallbackPtr;
    				/* The actual data. */

    acceptCallbackPtr = (AcceptCallback *) callbackData;
    if (acceptCallbackPtr->interp != (Tcl_Interp *) NULL) {
        UnregisterTcpServerInterpCleanup(acceptCallbackPtr->interp,
                acceptCallbackPtr);
    }
    Tcl_EventuallyFree((ClientData) acceptCallbackPtr->script, TCL_DYNAMIC);
    ckfree((char *) acceptCallbackPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SocketCmd --
 *
 *	This procedure is invoked to process the "socket" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Creates a socket based channel.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_SocketCmd(notUsed, interp, argc, argv)
    ClientData notUsed;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    int a, server, port, sockProtocol;
    char *arg, *copyScript, *host, *script;
    Tcl_Channel chan;
    AcceptCallback *acceptCallbackPtr;
    
    server = 0;
    script = NULL;

    if (TclHasSockets(interp) != TCL_OK) {
	return TCL_ERROR;
    }

    for (a = 1; a < argc; a++) {
        arg = argv[a];
	if (arg[0] == '-') {
	    if (strcmp(arg, "-server") == 0) {
		server = 1;
		a++;
		if (a >= argc) {
		    Tcl_AppendResult(interp, argv[0],
                            ": -server requires callback command",
                            (char *) NULL);
		    return TCL_ERROR;
		}
                script = argv[a];
            } else if (strcmp(arg, "-tcp") == 0) {
                sockProtocol = TCL_SOCKPROTO_TCP;
	    } else if (strcmp(arg, "--") == 0) {
		a++;
		break;
	    } else {
		Tcl_AppendResult(interp, argv[0], ": unknown option \"", arg,
                        "\"", (char *) NULL);
		return TCL_ERROR;
	    }
	} else {
	    break;
	}
    }
    if (server) {
        host = Tcl_GetHostName();
    } else if (a < argc) {
	host = argv[a];
        if (strcmp(host, "localhost") == 0) {
            host = Tcl_GetHostName();
        }
	a++;
    } else {
        Tcl_AppendResult(interp, argv[0], ": must specify host name",
                (char *) NULL);
        return TCL_ERROR;
    }

    if (a < argc) {
	if (TclSockGetPort(interp, argv[a], "tcp", &port) != TCL_OK) {
	    return TCL_ERROR;
	}
	a++;
    } else {
	Tcl_AppendResult(interp, argv[0], ": must specify port number",
                (char *) NULL);
        return TCL_ERROR;
    }
    if (a != argc) {
	Tcl_AppendResult(interp, argv[0], ": extra arguments: ", NULL);
	while (a < argc) {
	    Tcl_AppendResult(interp, argv[a], " ", (char *) NULL);
	    a++;
	}
	return TCL_ERROR;
    }
    
    if (server) {
        acceptCallbackPtr = (AcceptCallback *) ckalloc((unsigned)
                sizeof(AcceptCallback));
        copyScript = ckalloc((unsigned) strlen(script) + 1);
        strcpy(copyScript, script);
        acceptCallbackPtr->interp = interp;
        acceptCallbackPtr->script = copyScript;
        chan = Tcl_OpenTcpServer(interp, port, host, AcceptCallbackProc,
                (ClientData) acceptCallbackPtr);
        if (chan == (Tcl_Channel) NULL) {
            ckfree(copyScript);
            ckfree((char *) acceptCallbackPtr);
            return TCL_ERROR;
        }

        /*
         * Register with the interpreter to let us know when the
         * interpreter is deleted (by having the callback set the
         * acceptCallbackPtr->interp field to NULL). This is to
         * avoid trying to eval the script in a deleted interpreter.
         */

        RegisterTcpServerInterpCleanup(interp, acceptCallbackPtr);
        
        /*
         * Register a close callback. This callback will inform the
         * interpreter (if it still exists) that this channel does not
         * need to be informed when the interpreter is deleted.
         */
        
        Tcl_CreateCloseHandler(chan, TcpServerCloseCallback,
                (ClientData) acceptCallbackPtr);
    } else {
        chan = Tcl_OpenTcpClient(interp, port, host);
        if (chan == (Tcl_Channel) NULL) {
            return TCL_ERROR;
        }
    }
    Tcl_RegisterChannel(interp, chan);            
    Tcl_AppendResult(interp, Tcl_GetChannelName(chan), (char *) NULL);
    
    return TCL_OK;
}
