/* 
 * tclMacChan.c
 *
 *	Channel drivers for Macintosh channels based on
 *	TCP sockets, and the console fds.
 *
 * Copyright (c) 1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

static char sccsid[] = "@(#) tclMacChan.c 1.12 96/01/30 15:57:52";

#include "tclInt.h"
#include "tclPort.h"
#include <Aliases.h>
#include <Errors.h>
#include <Files.h>
#include <Gestalt.h>
#include <Processes.h>
#include <Strings.h>

/*
 * Static routines for this file:
 */

static int		UnixBlockMode _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile, int mode));
static int		CommonBufferSize _ANSI_ARGS_((ClientData instanceData));
static int		UnixInput _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, char *buf, int toRead,
			    int *errorCode));
static int		UnixOutput _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File outFile, char *buf, int toWrite,
			    int *errorCode));
static int		StdIOSeek _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile, long offset,
			    int mode, int *errorCode));
static Tcl_Channel	CreateStdFileChannel _ANSI_ARGS_((char *name,
			    Tcl_File inFile, Tcl_File outFile));
static int		StdIOClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile));
static void		TcpAccept _ANSI_ARGS_((ClientData data, int mask));
static int		TcpClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile));

/*
 * This structure describes the channel type structure for file based IO:
 */

static Tcl_ChannelType consoleChannelType = {
    "file",			/* Type name. */
    UnixBlockMode,		/* Set blocking or
                                 * non-blocking mode.*/
    CommonBufferSize,		/* Buffer size. */
    StdIOClose,			/* Close proc. */
    UnixInput,			/* Input proc. */
    UnixOutput,			/* Output proc. */
    StdIOSeek				/* Seek proc. */
};

/*
 * This structure describes the channel type structure for TCP socket
 * based IO:
 */

static Tcl_ChannelType tcpChannelType = {
    "tcp",			/* Type name. */
    UnixBlockMode,		/* Set blocking or
                                 * non-blocking mode.*/
    CommonBufferSize,		/* Buffer size. */
    TcpClose,			/* Close proc. */
    UnixInput,			/* Input proc. */
    UnixOutput,			/* Output proc. */
    NULL			/* Seek proc. */
};

/*
 * This is the size of the channel name for File based channels
 */

#define CHANNEL_NAME_SIZE	64
static char channelName[CHANNEL_NAME_SIZE+1];

/*
 * The preferred buffer size for Unix channels.
 */

#define CHANNEL_BUF_SIZE	4096

/*
 * This structure describes per-instance state of a tcp based channel.
 */

typedef struct TcpState {
    Tcl_File sock;			/* The socket itself. */
    Tcl_TcpAcceptProc *acceptProc;	/* Proc to call on accept. */
    ClientData acceptProcData;		/* The data for the accept proc. */
} TcpState;

/*
 * Static variables to hold channels for stdin, stdout and stderr.
 */

static Tcl_Channel stdinChannel = NULL;
static Tcl_Channel stdoutChannel = NULL;
static Tcl_Channel stderrChannel = NULL;

/*
 * Static variables to hold files for stdin, stdout and stderr.
 */

static Tcl_File stdinFile = NULL;
static Tcl_File stdoutFile = NULL;
static Tcl_File stderrFile = NULL;

/*
 * Flag used internally by socket driver to remember that the socket
 * is not yet connected.
 */

#define TCL_SOCKET_NOT_CONNECTED	(1<<2)

/*
 * Globals for holding information about OS support for sockets.
 */
static int socketsTestInited = false;
static int hasSockets = false;

/*
 *----------------------------------------------------------------------
 *
 * CommonBufferSize --
 *
 *	Returns the preferred buffer size for unix IO channels.
 *
 * Results:
 *	The preferred buffer size.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
CommonBufferSize(clientData)
    ClientData clientData;		/* Unused. */
{
    return CHANNEL_BUF_SIZE;
}

/*
 *----------------------------------------------------------------------
 *
 * UnixOutput--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
UnixOutput(instanceData, outFile, buf, toWrite, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File outFile;			/* Output device for channel. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    int written;
    int fd;

    *errorCode = 0;
    errno = 0;
    fd = (int) Tcl_GetFileInfo(outFile, NULL);
    written = write(fd, buf, (size_t) toWrite);
    if (written > -1) {
        return written;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * UnixInput --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

int
UnixInput(instanceData, inFile, buf, bufSize, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    struct stat statBuf;
    int fd;
    int bytesRead;			/* How many bytes were read? */

    *errorCode = 0;
    errno = 0;
    fd = (int) Tcl_GetFileInfo(inFile, NULL);
    if (fstat(fd, &statBuf) != 0) {
        *errorCode = errno;
        return -1;
    }
    if ((statBuf.st_size > 0) && (statBuf.st_size < bufSize)) {
        bufSize = statBuf.st_size;
    }
    bytesRead = read(fd, buf, (size_t) bufSize);
    if (bytesRead > -1) {
        return bytesRead;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * StdIOSeek --
 *
 *	Seeks on an IO channel. Returns the new position.
 *
 * Results:
 *	-1 if failed, the new position if successful. If failed, it
 *	also sets *errorCodePtr to the error code.
 *
 * Side effects:
 *	Moves the location at which the channel will be accessed in
 *	future operations.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
StdIOSeek(instanceData, inFile, outFile, offset, mode, errorCodePtr)
    ClientData instanceData;			/* Unused. */
    Tcl_File inFile, outFile;			/* Input and output
                                                 * devices for channel. */
    long offset;				/* Offset to seek to. */
    int mode;					/* Relative to where
                                                 * should we seek? */
    int *errorCodePtr;				/* To store error code. */
{
    int newLoc;
    int fd;

    *errorCodePtr = 0;
    if (inFile != (Tcl_File) NULL) {
        fd = (int) Tcl_GetFileInfo(inFile, NULL);
    } else if (outFile != (Tcl_File) NULL) {
        fd = (int) Tcl_GetFileInfo(outFile, NULL);
    } else {
        *errorCodePtr = EFAULT;
        return -1;
    }
    newLoc = lseek(fd, offset, mode);
    if (newLoc > -1) {
        return newLoc;
    }
    *errorCodePtr = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * UnixBlockMode --
 *
 *	Set blocking or non-blocking mode on channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
UnixBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile, outFile;		/* Input, output for channel. */
    int mode;				/* The mode to set. */
{
    int curStatus;
    int fd;

    if (stdinFile == (Tcl_File) NULL) {
        TclGetStdHandles(&stdinFile, &stdoutFile, &stderrFile);
    }

    /*
     * Do not allow putting stdin, stdout or stderr into nonblocking mode.
     */
    
    if (mode == TCL_MODE_NONBLOCKING) {
        if ((inFile == stdinFile) || (outFile == stdoutFile) ||
                (outFile == stderrFile)) {
            return EFAULT;
        }
    }
    
    if (inFile != NULL) {
        fd = (int) Tcl_GetFileInfo(inFile, NULL);
        curStatus = fcntl(fd, F_GETFL);
        if (mode == TCL_MODE_BLOCKING) {
            curStatus &= (~(O_NONBLOCK));
        } else {
            curStatus |= O_NONBLOCK;
        }
        if (fcntl(fd, F_SETFL, curStatus) < 0) {
            return errno;
        }
    }
    if (outFile != NULL) {
        fd = (int) Tcl_GetFileInfo(outFile, NULL);
        curStatus = fcntl(fd, F_GETFL);
        if (mode == TCL_MODE_BLOCKING) {
            curStatus &= (~(O_NONBLOCK));
        } else {
            curStatus |= O_NONBLOCK;
        }
        if (fcntl(fd, F_SETFL, curStatus) < 0) {
            return errno;
        }
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * StdIOClose --
 *
 *	Closes the IO channel.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the physical channel
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
StdIOClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* Unused. */
    Tcl_File inFile;		/* Input file to close. */
    Tcl_File outFile;		/* Output file to close. */
{
    int fd, errorCode = 0;

    /*
     * Invalidate the stdio cache if necessary.  Note that we assume that
     * the stdio file and channel pointers will become invalid at the same
     * time.
     */

    if (inFile == stdinFile) {
	stdinFile = NULL;
	stdinChannel = NULL;
    }
    if (outFile == stdoutFile) {
	stdoutFile = NULL;
	stdoutChannel = NULL;
    } else if (outFile == stderrFile) {
	stderrFile = NULL;
	stderrChannel = NULL;
    }

    if (inFile != NULL) {
        fd = (int) Tcl_GetFileInfo(inFile, NULL);

	/*
	 * Check for read/write file so we only close it once.
	 */

	if (inFile == outFile) {
	    outFile = NULL;
	}
        Tcl_FreeFile(inFile);

        if (fd > 2) {
            if (close(fd) < 0) {
		errorCode = errno;
	    }
        }
    }

    if (outFile != NULL) {
        fd = (int) Tcl_GetFileInfo(outFile, NULL);
        Tcl_FreeFile(outFile);        
        if (fd > 2) {
	    if ((close(fd) < 0) && (errorCode == 0)) {
		errorCode = errno;
	    }
        } 
    }
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetCommandPids --
 *
 *	Stores a list of the command PIDs for a command channel in
 *	interp->result.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Modifies interp->result.
 *
 *----------------------------------------------------------------------
 */

void
TclGetCommandPids(interp, chan)
    Tcl_Interp *interp;
    Tcl_Channel chan;
{
    return;
}

/*
 *----------------------------------------------------------------------
 *
 * TclCleanupCommandChannel --
 *
 *	Perform special cleanup for command channels.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May wait for child processes.
 *
 *----------------------------------------------------------------------
 */

int
TclCleanupCommandChannel(interp, chan, keepNewline)
    Tcl_Interp *interp;		/* Interpreter to report errors in. */
    Tcl_Channel chan;		/* Channel to clean up after. */
    int keepNewline;		/* non-zero means keep last newline
                                 * in interp->result. */
{
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpClose --
 *
 *	Close the socket.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* The socket to close. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    TcpState *statePtr = (TcpState *) instanceData;
    Tcl_File sock = statePtr->sock;
    int errorCode = 0;

    /*
     * We assume that inFile==outFile==sock and so only clean up one of them.
     */
    
    if (close((int) Tcl_GetFileInfo(sock, NULL)) < 0) {
	errorCode = errno;
    }
    
    /*
     * Delete a file handler that may be active for this socket.
     * Channel handlers are already deleted in the generic IO close
     * code which called this function.
     */
    
    Tcl_DeleteFileHandler(sock);

    ckfree((char *) statePtr);

    Tcl_FreeFile(sock);
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenCommandChannel --
 *
 *	Opens a communication channel to a sequence of commands.
 *	Command channels are not supported on the Macintosh.  This
 *	will alway return an NULL channel.
 *
 * Results:
 *	NULL.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
Tcl_Channel
Tcl_OpenCommandChannel(interp, argc, argv, modeString)
    Tcl_Interp *interp;		/* Interpreter for error reporting. */
    int argc;			/* How many arguments. */
    char **argv;		/* Array of arguments for command pipe. */
    char *modeString;		/* Which modes to open the channel in? */
{
    Tcl_AppendResult(interp, "pipes not supported on Macintosh OS", NULL);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_PidCmd --
 *
 *	This procedure is invoked to process the "pid" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_PidCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    ProcessSerialNumber psn;
    Tcl_Channel chan;

    if (argc > 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " ?channelId?\"", (char *) NULL);
	return TCL_ERROR;
    }
    
    if (argc == 2) {
        chan = Tcl_GetChannel(interp, argv[1], NULL);
	if (chan == (Tcl_Channel) NULL) {
	    return TCL_ERROR;
	}
	
	/*
	 * We can't create pipelines on the Mac so
	 * this will always return an empty list.
	 */
	return TCL_OK;
    }
    
    GetCurrentProcess(&psn);
    sprintf(interp->result, "0x%08x%08x", psn.highLongOfPSN, psn.lowLongOfPSN);
    
    return TCL_OK;
    }

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpClient --
 *
 *	Opens a TCP client socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed. On failure, the routine also
 *	sets the output argument errorCodePtr to the error code.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpClient(interp, port, host)
    Tcl_Interp *interp;			/* For error reporting; can be NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Host on which to open port. */
{
    struct sockaddr_in sockaddr;	/* Socket address */
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */
    struct linger ling;			/* For making sockets linger. */
    Tcl_File handle;			/* Socket file handle. */
    int sock;				/* Socket file descriptor */
    int status;
    Tcl_Channel chan;
    TcpState *sockState;

    if (TclHasSockets(interp) != TCL_OK) {
	return NULL;
    }

    (void) memset((char *) &sockaddr, '\0', sizeof(struct sockaddr_in));
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_port = (short) (port & 0xFFFF);
    hostent = gethostbyname(host);
    if (hostent != NULL) {
	memcpy((char *) &addr,
	    (char *) hostent->h_addr_list[0], (size_t) hostent->h_length);
    } else {
	addr = inet_addr(host);
	if (addr.s_addr == (unsigned long)-1) {
            Tcl_SetErrno(errno);
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "couldn't open socket: ",
                        Tcl_PosixError(interp), (char *) NULL);
            }
	    return (Tcl_Channel) NULL;
	}
    }

    /*
     * There is a rumor that this assignment may require care on
     * some 64 bit machines.
     */

    sockaddr.sin_addr.s_addr = addr.s_addr;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
	return (Tcl_Channel) NULL;
    }

    /*
     * Make server socket close gracefully.
     */
    
    ling.l_onoff = 1;
    ling.l_linger = 0;
    (void) setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &ling,
            sizeof(ling));
        
    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, CHANNEL_BUF_SIZE);

    handle = Tcl_GetFile((ClientData) sock, TCL_UNIX_FD);

    sockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    sockState->sock = handle;
    sockState->acceptProc = (Tcl_TcpAcceptProc *) NULL;
    sockState->acceptProcData = (ClientData) NULL;

    /*
     * Attempt to connect. The connect may fail at present with an
     * EINPROGRESS but at a later time it will complete. The caller
     * will set up a file handler on the socket if she is interested in
     * being informed when the connect completes.
     */
    
    status = connect(sock, (struct sockaddr *) &sockaddr, sizeof(sockaddr));
    if (status < 0) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        close(sock);
        Tcl_FreeFile(handle);
        ckfree((char *) sockState);
        return (Tcl_Channel) NULL;
    }

    sprintf(channelName, "sock%d", (int) sock);
    chan = Tcl_CreateChannel(&tcpChannelType, channelName, handle, handle,
            (ClientData) sockState);
    if (chan == (Tcl_Channel) NULL) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        close(sock);
        Tcl_FreeFile(handle);
        ckfree((char *) sockState);
        return NULL;
    }
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpServer --
 *
 *	Opens a TCP server socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.
 *
 * Side effects:
 *	Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpServer(interp, port, host, acceptProc, acceptProcData)
    Tcl_Interp *interp;			/* For error reporting - may be
                                         * NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Name of local host. */
    Tcl_TcpAcceptProc *acceptProc;	/* Callback for accepting connections
                                         * from new clients. */
    ClientData acceptProcData;		/* Data for the callback. */
{
    struct sockaddr_in sockaddr;	/* Socket address */
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */
    struct linger ling;			/* For making sockets linger. */
    Tcl_File handle;			/* Socket file handle. */
    int sock;				/* Socket file descriptor */
    int status;
    TcpState *sockState;
    Tcl_Channel chan;

    if (TclHasSockets(interp) != TCL_OK) {
	return NULL;
    }

    (void) memset((char *) &sockaddr, '\0', sizeof(struct sockaddr_in));
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_port = (short) (port & 0xFFFF);
    hostent = gethostbyname(host);
    if (hostent != NULL) {
	memcpy((char *) &addr,
	    (char *) hostent->h_addr_list[0], (size_t) hostent->h_length);
    } else {
	addr = inet_addr(host);
	if (addr.s_addr == (unsigned long)-1) {
            Tcl_SetErrno(errno);
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "couldn't open socket: ",
                        Tcl_PosixError(interp), (char *) NULL);
            }
            return (Tcl_Channel) NULL;
	}
    }

    /*
     * There is a rumor that this assignment may require care on
     * some 64 bit machines.
     */

    sockaddr.sin_addr.s_addr = addr.s_addr;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        return (Tcl_Channel) NULL;
    }

    /*
     * Set up to reuse server addresses automatically.
     */
    
    status = 1;
    (void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
            sizeof(status));

    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, CHANNEL_BUF_SIZE);

    /*
     * Bind to the specified port.
     */
    
    status = bind(sock, (struct sockaddr *) &sockaddr, sizeof(sockaddr));
    if (status < 0) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        close(sock);
        return (Tcl_Channel) NULL;
    }

    /*
     * Listen for connections.
     */

    listen(sock, 5);

    /*
     * Make server socket close down gracefully.
     */
    
    ling.l_onoff = 1;
    ling.l_linger = 0;
    (void) setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &ling,
            sizeof(ling));

    handle = Tcl_GetFile((ClientData) sock, TCL_UNIX_FD);

    sockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    sockState->sock = handle;
    sockState->acceptProc = acceptProc;
    sockState->acceptProcData = acceptProcData;

    /*
     * Set up the callback mechanism for accepting connections
     * from new clients. The caller will use Tcl_TcpRegisterCallback
     * to register a callback to call when a new connection is
     * accepted.
     */

    Tcl_CreateFileHandler(handle, TCL_READABLE, TcpAccept,
            (ClientData) sockState);
    
    sprintf(channelName, "sock%d", (int) sock);
    chan = Tcl_CreateChannel(&tcpChannelType, channelName, NULL, NULL,
            (ClientData) sockState);
    if (chan == (Tcl_Channel) NULL) {
        Tcl_SetErrno(errno);
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        ckfree((char *) sockState);
        close(sock);
        Tcl_DeleteFileHandler(handle);
        Tcl_FreeFile(handle);
        return NULL;
    }
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpAccept --
 *	Accept a TCP socket connection.  This is called by the event loop,
 *	and it in turns calls any registered callbacks for this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Evals the Tcl script associated with the server socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
TcpAccept(data, mask)
    ClientData data;			/* Callback token. */
    int mask;				/* Not used. */
{
    TcpState *sockState;		/* Client data of server socket. */
    int newsock;			/* The new client socket */
    Tcl_File newFile;			/* Its file. */
    TcpState *newSockState;		/* State for new socket. */
    struct sockaddr_in addr;		/* The remote address */
    int len;				/* For accept interface */
    Tcl_Channel  chan;			/* Channel instance created. */

    sockState = (TcpState *) data;

    len = sizeof(struct sockaddr_in);
    newsock = accept((int) Tcl_GetFileInfo(sockState->sock, NULL),
	    (struct sockaddr *)&addr, &len);
    if (newsock < 0) {
        return;
    }
    
    newFile = Tcl_GetFile((ClientData) newsock, TCL_UNIX_FD);
    if (newFile) {
        newSockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));

        newSockState->sock = newFile;
        newSockState->acceptProc = (Tcl_TcpAcceptProc *) NULL;
        newSockState->acceptProcData = (ClientData) NULL;
        
        sprintf(channelName, "sock%d", (int) newsock);
        chan = Tcl_CreateChannel(&tcpChannelType, channelName, newFile,
                newFile, (ClientData) newSockState);
        if (chan == (Tcl_Channel) NULL) {
            ckfree((char *) newSockState);
            close(newsock);
            Tcl_FreeFile(newFile);
	} else {
            if (sockState->acceptProc != (Tcl_TcpAcceptProc *) NULL) {
                (sockState->acceptProc) (sockState->acceptProcData, chan,
                        inet_ntoa(addr.sin_addr));
            }
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetStdChannels --
 *
 *	Returns handles to the standard channels for stdin, stdout and
 *	stderr.
 *
 * Results:
 *	None; returns values through output parameters.
 *
 * Side effects:
 *	May cause the creation of these standard channels and the underlying
 *	files.
 *
 *----------------------------------------------------------------------
 */

void
TclGetStdChannels(stdinPtr, stdoutPtr, stderrPtr)
    Tcl_Channel *stdinPtr,		/* Where to store the channels. */
    	        *stdoutPtr,
    		*stderrPtr;
{
    /*
     * If the channels were not created yet, create them now and
     * store them in the static variables.
     */

    if (stdinPtr != NULL) {
	if (stdinFile == NULL) {
	    stdinFile = Tcl_GetFile((ClientData) 0, TCL_UNIX_FD);
	    stdinChannel = CreateStdFileChannel("stdin", stdinFile, NULL);
        }
        *stdinPtr = (Tcl_Channel) stdinChannel;
    }
    if (stdoutPtr != NULL) {
	if (stdoutFile == NULL) {
	    stdoutFile = Tcl_GetFile((ClientData) 1, TCL_UNIX_FD);
	    stdoutChannel = CreateStdFileChannel("stdout", NULL, stdoutFile);
        }
	*stdoutPtr = (Tcl_Channel) stdoutChannel;
    }
    if (stderrPtr != NULL) {
        if (stderrFile == NULL) {
	    stderrFile = Tcl_GetFile((ClientData) 2, TCL_UNIX_FD);
	    stderrChannel = CreateStdFileChannel("stderr", NULL, stderrFile);
        }
        *stderrPtr = (Tcl_Channel) stderrChannel;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetHostName --
 *
 *	Returns the name of the local host.
 *
 * Results:
 *	Returns a string containing the host name, or NULL on error.
 *	The returned string must be freed by the caller.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static char hostname[255];	/* This buffer should be big enough for
                                 * hostname plus domain name. */
static int  hostnameInited = 0;

char *
Tcl_GetHostName()
{
    if (hostnameInited) {
        return hostname;
    }
    
    if (TclHasSockets(NULL) != TCL_OK) {
	hostname[0] = '\0';
        hostnameInited = 1;
	return hostname;
    }

    if (gethostname(hostname, 100) == 0) {
        hostnameInited = 1;
        return hostname;
    }
    return (char *) NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateStdFileChannel --
 *
 *	Create a channel suitable for use in wrapping a standard
 *	file handle.  Sets up the channel for use by the file driver.
 *
 * Results:
 *	Returns a newly allocated file channel.
 *
 * Side effects:
 *	Registers a new channel with the given name.
 *
 *----------------------------------------------------------------------
 */

static Tcl_Channel
CreateStdFileChannel(name, inFile, outFile)
    char *name;			/* Name of standard channel. */
    Tcl_File inFile;		/* File to use for input, or NULL. */
    Tcl_File outFile;		/* File to use for output, or NULL. */
{
    Tcl_Channel channel;

    channel = Tcl_CreateChannel(&consoleChannelType, name, inFile, outFile,
	    (ClientData) NULL);
    /*
     * Set up the normal channel options for stdio handles.
     */

    if (channel != NULL) {
	Tcl_SetChannelOption(NULL, channel, "-translation", "cr");
	Tcl_SetChannelOption(NULL, channel, "-linemode", "on");
    }
    return channel;
}

/*
 *----------------------------------------------------------------------
 *
 * TclHasSockets --
 *
 *	This function determines whether sockets are available on the
 *	current system and returns an error in interp if they are not.
 *	Note that interp may be NULL.  This call uses the Macintosh
 *	gestalt function to determine the existance of Mac Tcp.
 *
 * Results:
 *	Returns TCL_OK if the system supports sockets, or TCL_ERROR with
 *	an error in interp.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

#define gestaltMacTCPVersion 'mtcp'
int
TclHasSockets(interp)
    Tcl_Interp *interp;
{
    long response;
    
static int socketsTestInited = false;
static int hasSockets = false;
    if (!socketsTestInited) {
	if (Gestalt(gestaltMacTCPVersion, &response) == noErr) {
	    hasSockets = true;
	} else {
	    hasSockets = false;
	}
	socketsTestInited = true;
    }
    if (hasSockets) {
	return TCL_OK;
    }
    if (interp != NULL) {
	Tcl_AppendResult(interp, "sockets are not available on this system",
		NULL);
    }
    return TCL_ERROR;
}
