/* 
 * tclUnixChan.c
 *
 *	Common channel driver for Unix channels based on files, command
 *	pipes and TCP sockets.
 *
 * Copyright (c) 1995-1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

static char sccsid[] = "@(#) tclUnixChan.c 1.111 96/01/19 18:20:21";

#include	"tclInt.h"	/* Internal definitions for Tcl. */
#include	"tclPort.h"	/* Portability features for Tcl. */

/*
 * This structure describes per-instance state of a pipe based channel.
 */

typedef struct PipeState {
    Tcl_File inPipe;	/* Output from pipe. */
    Tcl_File outPipe;	/* Input to pipe. */
    Tcl_File errorId;	/* Error output from pipe. */
    int numPids;	/* How many processes are attached to this pipe? */
    int *pidPtr;	/* The process IDs themselves. */
} PipeState;

/*
 * This structure describes per-instance state of a tcp based channel.
 */

typedef struct TcpState {
    Tcl_File sock;			/* The socket itself. */
    Tcl_TcpAcceptProc *acceptProc;	/* Proc to call on accept. */
    ClientData acceptProcData;		/* The data for the accept proc. */
} TcpState;

/*
 * Static routines for this file:
 */

static int		CommonBlockMode _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile, int mode));
static int		CommonBufferSize _ANSI_ARGS_((
    			    ClientData instanceData));
static int		FilePipeInput _ANSI_ARGS_((ClientData instanceData,
		            Tcl_File inFile, char *buf, int toRead,
		            int *errorCode));
static int		CommonOutput _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File outFile, char *buf, int toWrite,
			    int *errorCode));
static TcpState *	CreateSocket _ANSI_ARGS_((Tcl_Interp *interp,
			    int port, char *host, int server));
static Tcl_Channel	CreateStdFileChannel _ANSI_ARGS_((char *name,
			    Tcl_File inFile, Tcl_File outFile));
static int		FileClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile));
static int		FileSeek _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile, long offset,
			    int mode, int *errorCode));
static int		PipeClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile));
static void		TcpAccept _ANSI_ARGS_((ClientData data, int mask));
static int		TcpClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile));
static int		TcpInput _ANSI_ARGS_((ClientData instanceData,
		            Tcl_File infile, char *buf, int toRead,
        		    int *errorCode));

/*
 * This structure describes the channel type structure for file based IO:
 */

static Tcl_ChannelType fileChannelType = {
    "file",				/* Type name. */
    CommonBlockMode,			/* Set blocking/nonblocking mode.*/
    CommonBufferSize,			/* Buffer size. */
    FileClose,				/* Close proc. */
    FilePipeInput,			/* Input proc. */
    CommonOutput,			/* Output proc. */
    FileSeek				/* Seek proc. */
};

/*
 * This structure describes the channel type structure for command pipe
 * based IO:
 */

static Tcl_ChannelType pipeChannelType = {
    "pipe",				/* Type name. */
    CommonBlockMode,			/* Set blocking/nonblocking mode.*/
    CommonBufferSize,			/* Buffer size. */
    PipeClose,				/* Close proc. */
    FilePipeInput,			/* Input proc. */
    CommonOutput,			/* Output proc. */
    NULL				/* Seek proc. */
};

/*
 * This structure describes the channel type structure for TCP socket
 * based IO:
 */

static Tcl_ChannelType tcpChannelType = {
    "tcp",				/* Type name. */
    CommonBlockMode,			/* Set blocking/nonblocking mode.*/
    CommonBufferSize,			/* Buffer size. */
    TcpClose,				/* Close proc. */
    TcpInput,				/* Input proc. */
    CommonOutput,			/* Output proc. */
    NULL				/* Seek proc. */
};

/*
 * The preferred buffer size for Unix channels.
 */

#define CHANNEL_BUF_SIZE	4096

/*
 * Static variables to hold channels for stdin, stdout and stderr.
 */

static Tcl_Channel stdinChannel = NULL;
static Tcl_Channel stdoutChannel = NULL;
static Tcl_Channel stderrChannel = NULL;

/*
 * Static variables to hold files for stdin, stdout and stderr.
 */

static Tcl_File stdinFile = NULL;
static Tcl_File stdoutFile = NULL;
static Tcl_File stderrFile = NULL;

/*
 *----------------------------------------------------------------------
 *
 * CommonBufferSize --
 *
 *	Returns the preferred buffer size for unix IO channels.
 *
 * Results:
 *	The preferred buffer size.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
CommonBufferSize(clientData)
    ClientData clientData;		/* Unused. */
{
    return CHANNEL_BUF_SIZE;
}

/*
 *----------------------------------------------------------------------
 *
 * FilePipeInput --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *	This routine is used by the file and command pipeline drivers.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
FilePipeInput(instanceData, inFile, buf, toRead, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int toRead;				/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    int fd;				/* The OS handle for reading. */
    int bytesRead;			/* How many bytes were actually
                                         * read from the input device? */

    *errorCode = 0;
    errno = 0;
    fd = (int) Tcl_GetFileInfo(inFile, NULL);
    
    /*
     * Assume there is always enough input available. This will block
     * appropriately, and read will unblock as soon as a short read is
     * possible, if the channel is in blocking mode. If the channel is
     * nonblocking, the read will never block.
     */
    
    if (toRead == 0) {
        return 0;
    }
    bytesRead = read(fd, buf, (size_t) toRead);
    if (bytesRead > -1) {
        return bytesRead;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * CommonOutput--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
CommonOutput(instanceData, outFile, buf, toWrite, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File outFile;			/* Output device for channel. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    int written;
    int fd;

    *errorCode = 0;
    errno = 0;
    fd = (int) Tcl_GetFileInfo(outFile, NULL);
    written = write(fd, buf, (size_t) toWrite);
    if (written > -1) {
        return written;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * CommonBlockMode --
 *
 *	Set blocking or non-blocking mode on channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
CommonBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile, outFile;		/* Input, output for channel. */
    int mode;				/* The mode to set. */
{
    int curStatus;
    int fd;

    if (stdinFile == (Tcl_File) NULL) {
        TclGetStdHandles(&stdinFile, &stdoutFile, &stderrFile);
    }

    /*
     * Do not allow putting stdin, stdout or stderr into nonblocking mode.
     */
    
    if (mode == TCL_MODE_NONBLOCKING) {
        if ((inFile == stdinFile) || (outFile == stdoutFile) ||
                (outFile == stderrFile)) {
            return EFAULT;
        }
    }
    
    if (inFile != NULL) {
        fd = (int) Tcl_GetFileInfo(inFile, NULL);
        curStatus = fcntl(fd, F_GETFL);
        if (mode == TCL_MODE_BLOCKING) {
            curStatus &= (~(O_NONBLOCK));
        } else {
            curStatus |= O_NONBLOCK;
        }
        if (fcntl(fd, F_SETFL, curStatus) < 0) {
            return errno;
        }
        curStatus = fcntl(fd, F_GETFL);
    }
    if (outFile != NULL) {
        fd = (int) Tcl_GetFileInfo(outFile, NULL);
        curStatus = fcntl(fd, F_GETFL);
        if (mode == TCL_MODE_BLOCKING) {
            curStatus &= (~(O_NONBLOCK));
        } else {
            curStatus |= O_NONBLOCK;
        }
        if (fcntl(fd, F_SETFL, curStatus) < 0) {
            return errno;
        }
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * FileClose --
 *
 *	Closes the IO channel.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the physical channel
 *
 *----------------------------------------------------------------------
 */

static int
FileClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* Unused. */
    Tcl_File inFile;		/* Input file to close. */
    Tcl_File outFile;		/* Output file to close. */
{
    int fd, errorCode = 0;

    /*
     * Invalidate the stdio cache if necessary.  Note that we assume that
     * the stdio file and channel pointers will become invalid at the same
     * time.
     */

    if (inFile == stdinFile) {
	stdinFile = NULL;
	stdinChannel = NULL;
    }
    if (outFile == stdoutFile) {
	stdoutFile = NULL;
	stdoutChannel = NULL;
    } else if (outFile == stderrFile) {
	stderrFile = NULL;
	stderrChannel = NULL;
    }

    if (inFile != NULL) {
        fd = (int) Tcl_GetFileInfo(inFile, NULL);

	/*
	 * Check for read/write file so we only close it once.
	 */

	if (inFile == outFile) {
	    outFile = NULL;
	}
        Tcl_FreeFile(inFile);

        if (fd > 2) {
            if (close(fd) < 0) {
		errorCode = errno;
	    }
        }
    }

    if (outFile != NULL) {
        fd = (int) Tcl_GetFileInfo(outFile, NULL);
        Tcl_FreeFile(outFile);        
        if (fd > 2) {
	    if ((close(fd) < 0) && (errorCode == 0)) {
		errorCode = errno;
	    }
        } 
    }
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * FileSeek --
 *
 *	Seeks on a file-based IO channel. Returns the new position.
 *
 * Results:
 *	-1 if failed, the new position if successful. If failed, it
 *	also sets *errorCodePtr to the error code.
 *
 * Side effects:
 *	Moves the location at which the channel will be accessed in
 *	future operations.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
FileSeek(instanceData, inFile, outFile, offset, mode, errorCodePtr)
    ClientData instanceData;			/* Unused. */
    Tcl_File inFile, outFile;			/* Input and output
                                                 * devices for channel. */
    long offset;				/* Offset to seek to. */
    int mode;					/* Relative to where
                                                 * should we seek? */
    int *errorCodePtr;				/* To store error code. */
{
    int newLoc;
    int fd;

    *errorCodePtr = 0;
    if (inFile != (Tcl_File) NULL) {
        fd = (int) Tcl_GetFileInfo(inFile, NULL);
    } else if (outFile != (Tcl_File) NULL) {
        fd = (int) Tcl_GetFileInfo(outFile, NULL);
    } else {
        *errorCodePtr = EFAULT;
        return -1;
    }
    newLoc = lseek(fd, offset, mode);
    if (newLoc > -1) {
        return newLoc;
    }
    *errorCodePtr = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpInput --
 *
 *	Reads input from the TCP IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpInput(instanceData, inFile, buf, bufSize, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    int sock;				/* The OS handle. */
    int bytesRead;			/* How many bytes were read? */

    *errorCode = 0;
    errno = 0;
    sock = (int) Tcl_GetFileInfo(inFile, NULL);

    if (bufSize == 0) {
        return 0;
    }
    bytesRead = recv(sock, buf, bufSize, 0);
    if (bytesRead > -1) {
        return bytesRead;
    }
    if (errno == ECONNRESET) {

        /*
         * Turn ECONNRESET into a soft EOF condition.
         */
        
        return 0;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetCommandPids --
 *
 *	Stores a list of the command PIDs for a command channel in
 *	interp->result.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Modifies interp->result.
 *
 *----------------------------------------------------------------------
 */

void
TclGetCommandPids(interp, chan)
    Tcl_Interp *interp;
    Tcl_Channel chan;
{
    PipeState *pipePtr;
    Tcl_ChannelType *chanTypePtr;
    int i;
    char buf[20];

    /*
     * Punt if the channel is not a command channel.
     */

    chanTypePtr = Tcl_GetChannelType(chan);
    if (chanTypePtr != &pipeChannelType) {
        return;
    }

    pipePtr = (PipeState *) Tcl_GetChannelInstanceData(chan);
    for (i = 0; i < pipePtr->numPids; i++) {
        sprintf(buf, "%d", pipePtr->pidPtr[i]);
        Tcl_AppendElement(interp, buf);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclCleanupCommandChannel --
 *
 *	Perform special cleanup for command channels.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May wait for child processes.
 *
 *----------------------------------------------------------------------
 */

int
TclCleanupCommandChannel(interp, chan, keepNewline)
    Tcl_Interp *interp;		/* Interpreter to report errors in. */
    Tcl_Channel chan;		/* Channel to clean up after. */
    int keepNewline;		/* non-zero means keep last newline
                                 * in interp->result. */
{
    int result;
    PipeState *pipePtr;
    Tcl_ChannelType *chanTypePtr;
    Tcl_Channel errChan;
    int fd;
    
    /*
     * Punt if the channel is not a command channel.
     */

    chanTypePtr = Tcl_GetChannelType(chan);
    if (chanTypePtr != &pipeChannelType) {
        return TCL_OK;
    }

    pipePtr = (PipeState *) Tcl_GetChannelInstanceData(chan);
    if (pipePtr->inPipe != NULL) {
        fd = (int) Tcl_GetFileInfo(pipePtr->inPipe, NULL);
        close(fd);
        Tcl_FreeFile(pipePtr->inPipe);
        pipePtr->inPipe = (Tcl_File) NULL;
    }
    if (pipePtr->outPipe != NULL) {
        fd = (int) Tcl_GetFileInfo(pipePtr->outPipe, NULL);
        close(fd);
        Tcl_FreeFile(pipePtr->outPipe);
        pipePtr->outPipe = (Tcl_File) NULL;
    }
    
    /*
     * Wrap the error file into a channel and give it to the cleanup
     * routine.
     */

    if (pipePtr->errorId != NULL) {
	errChan = Tcl_CreateChannel(&fileChannelType, NULL, pipePtr->errorId,
		NULL, NULL);
    } else {
	errChan = NULL;
    }
    result = TclCleanupChildren(interp, pipePtr->numPids, pipePtr->pidPtr,
            errChan, keepNewline);
    
    pipePtr->numPids = 0;
    pipePtr->errorId = (Tcl_File) NULL;
    
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * PipeClose --
 *
 *	Closes a pipe based IO channel.
 *
 * Results:
 *	0 on success, errno otherwise.
 *
 * Side effects:
 *	Closes the physical channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
PipeClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* The pipe to close. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    PipeState *pipePtr;
    int fd, i, errorCode = 0;

    pipePtr = (PipeState *) instanceData;
    if (pipePtr->inPipe != NULL) {
        fd = (int) Tcl_GetFileInfo(pipePtr->inPipe, NULL);
        Tcl_FreeFile(pipePtr->inPipe);
	if (close(fd) < 0) {
	    errorCode = errno;
	}
    }
    if (pipePtr->outPipe != NULL) {
        fd = (int) Tcl_GetFileInfo(pipePtr->outPipe, NULL);
        Tcl_FreeFile(pipePtr->outPipe);
	if ((close(fd) < 0) && (errorCode == 0)) {
	    errorCode = errno;
	}
    }
    if (pipePtr->errorId != NULL) {
        fd = (int) Tcl_GetFileInfo(pipePtr->errorId, NULL);
        Tcl_FreeFile(pipePtr->errorId);
	if ((close(fd) < 0) && (errorCode == 0)) {
	    errorCode = errno;
	}
    }
    if (pipePtr->numPids > 0) {
        for (i = 0; i < pipePtr->numPids; i++) {
            if (pipePtr->pidPtr[i] != -1) {
                Tcl_DetachPids(1, &pipePtr->pidPtr[i]);
            }
        }
        ckfree((char *) pipePtr->pidPtr);
    }
    ckfree((char *) pipePtr);
        
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpClose --
 *
 *	Close the socket.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TcpClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* The socket to close. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    TcpState *statePtr;
    Tcl_File sockFile;
    int sock;
    int errorCode = 0;

    statePtr = (TcpState *) instanceData;
    sockFile = statePtr->sock;
    sock = (int) Tcl_GetFileInfo(sockFile, NULL);
    
    /*
     * Delete a file handler that may be active for this socket.
     * Channel handlers are already deleted in the generic IO close
     * code which called this function.
     */
    
    Tcl_DeleteFileHandler(sockFile);

    ckfree((char *) statePtr);
    
    /*
     * We assume that inFile==outFile==sockFile and so
     * we only clean up sockFile.
     */

    Tcl_FreeFile(sockFile);

    if (close(sock) < 0) {
	errorCode = errno;
    }

    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenFileChannel --
 *
 *	Open an File based channel on Unix systems.
 *
 * Results:
 *	The new channel or NULL. If NULL, the output argument
 *	errorCodePtr is set to a POSIX error.
 *
 * Side effects:
 *	May open the channel and may cause creation of a file on the
 *	file system.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenFileChannel(interp, fileName, modeString, permissions)
    Tcl_Interp *interp;			/* Interpreter for error reporting;
                                         * can be NULL. */
    char *fileName;			/* Name of file to open. */
    char *modeString;			/* A list of POSIX open modes or
                                         * a string such as "rw". */
    int permissions;			/* If the open involves creating a
                                         * file, with what modes to create
                                         * it? */
{
    int fd, seekFlag, mode, channelPermissions;
    Tcl_File file;
    Tcl_Channel chan;
    char *nativeName, channelName[20];
    Tcl_DString buffer;

    mode = TclGetOpenMode(interp, modeString, &seekFlag);
    if (mode == -1) {
        return NULL;
    }
    switch (mode & (O_RDONLY | O_WRONLY | O_RDWR)) {
	case O_RDONLY:
	    channelPermissions = TCL_READABLE;
	    break;
	case O_WRONLY:
	    channelPermissions = TCL_WRITABLE;
	    break;
	case O_RDWR:
	    channelPermissions = (TCL_READABLE | TCL_WRITABLE);
	    break;
	default:
	    panic("Tcl_OpenFileChannel: invalid mode value");
	    return NULL;
    }

    nativeName = Tcl_TranslateFileName(interp, fileName, &buffer);
    if (nativeName == NULL) {
	return NULL;
    }
    fd = open(nativeName, mode, permissions);
    Tcl_DStringFree(&buffer);

    if (fd < 0) {
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open \"", fileName, "\": ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        return NULL;
    }
    
    file = Tcl_GetFile((ClientData) fd, TCL_UNIX_FD);
    sprintf(channelName, "file%d", (int) Tcl_GetFileInfo(file, NULL));
    
    chan = Tcl_CreateChannel(&fileChannelType, channelName,
            (channelPermissions & TCL_READABLE) ? file : NULL,
            (channelPermissions & TCL_WRITABLE) ? file : NULL,
            (ClientData) NULL);
    if (chan == (Tcl_Channel) NULL) {
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "could not open channel \"",
                    channelName, "\": ", Tcl_PosixError(interp),
                    (char *) NULL);
        }
        Tcl_FreeFile(file);
        close(fd);
        return NULL;
    }

    if (seekFlag) {
        if (Tcl_Seek(chan, 0, SEEK_END) < 0) {
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "could not seek to end of file on \"",
                        channelName, "\": ", Tcl_PosixError(interp),
                        (char *) NULL);
            }
            Tcl_Close(chan);
            return NULL;
        }
    }
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenCommandChannel --
 *
 *	Opens a communication channel to a sequence of commands. 
 *
 * Results:
 *	The channel or NULL on failure.
 *
 * Side effects:
 *	Creates processes, opens pipes.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
Tcl_Channel
Tcl_OpenCommandChannel(interp, argc, argv, modeString)
    Tcl_Interp *interp;		/* Interpreter for error reporting. Can
                                 * NOT be NULL. */
    int argc;			/* How many arguments. */
    char **argv;		/* Array of arguments for command pipe. */
    char *modeString;		/* Which modes to open the channel in? */
{
    int permissions = 0;	/* Which end(s) of pipe to read, write. */
    PipeState *pipePtr;		/* The instance data. */
    Tcl_File *inPipePtr, *outPipePtr;
    Tcl_File inPipe, outPipe, errorId, pipeId;
    int fd, numPids, *pidPtr, channelId;
    int mode, seekFlag;
    Tcl_Channel chan;
    char channelName[20];

    inPipe = outPipe = errorId = NULL;

    /*
     * If the modeString is NULL, then we want either a readable or a null
     * channel.  This is used by exec; it can't tell what mode to request
     * because there may be redirections.  Otherwise we parse the mode
     * string and set the permissions appropriately.
     */

    if (modeString == NULL) {
	outPipePtr = &outPipe;
	inPipePtr = NULL;
    } else {
        mode = TclGetOpenMode(interp, modeString, &seekFlag);
        if (mode == -1) {
            return (Tcl_Channel) NULL;
        }
	switch (mode & (O_RDONLY | O_WRONLY | O_RDWR)) {
	    case O_RDONLY:
		permissions = TCL_READABLE;
		break;
	    case O_WRONLY:
		permissions = TCL_WRITABLE;
		break;
	    case O_RDWR:
		permissions = (TCL_READABLE | TCL_WRITABLE);
		break;
	    default:
		panic("Tcl_OpenCommandChannel: invalid mode value");
		break;
	}
	inPipePtr = (permissions & TCL_WRITABLE) ? &inPipe : NULL;
	outPipePtr = (permissions & TCL_READABLE) ? &outPipe : NULL;
    }
    
    numPids = Tcl_CreatePipeline(interp, argc, argv, &pidPtr, inPipePtr,
            outPipePtr, &errorId);

    if (numPids < 0) {
        if (inPipe != NULL) {
            fd = (int) Tcl_GetFileInfo(inPipe, NULL);
            close(fd);
            Tcl_FreeFile(inPipe);
        }
        if (outPipe != NULL) {
            fd = (int) Tcl_GetFileInfo(outPipe, NULL);
            close(fd);
            Tcl_FreeFile(outPipe);
        }
        if (errorId != NULL) {
            fd = (int) Tcl_GetFileInfo(errorId, NULL);
            close(fd);
	    Tcl_FreeFile(errorId);
        }
        return NULL;
    }

    if (modeString != NULL) {
	
	/*
	 * Since we were called with a mode string, we need to
	 * verify that the permissions match the set of pipes
	 * that were created.
	 */

	if (permissions & TCL_READABLE) {
	    if (outPipe == NULL) {
		if (inPipe != NULL) {
		    fd = (int) Tcl_GetFileInfo(inPipe, NULL);
		    close(fd);
		    Tcl_FreeFile(inPipe);
		}
		Tcl_AppendResult(interp, "can't read output from command:",
			" standard output was redirected", (char *) NULL);
		if (numPids > 0) {
		    Tcl_DetachPids(numPids, pidPtr);
		    ckfree((char *) pidPtr);
		}
		if (errorId != NULL) {
		    fd = (int) Tcl_GetFileInfo(errorId, NULL);
		    close(fd);
		    Tcl_FreeFile(errorId);
		}
		return NULL;
	    }
	}
	if (permissions & TCL_WRITABLE) {
	    if (inPipe == NULL) {
		if (outPipe != NULL) {
		    fd = (int) Tcl_GetFileInfo(outPipe, NULL);
		    close(fd);
		    Tcl_FreeFile(outPipe);
		}
		Tcl_AppendResult(interp, "can't write input to command:",
			" standard input was redirected", (char *) NULL);
		if (numPids > 0) {
		    Tcl_DetachPids(numPids, pidPtr);
		    ckfree((char *) pidPtr);
		}
		if (errorId != NULL) {
		    fd = (int) Tcl_GetFileInfo(errorId, NULL);
		    close(fd);
		    Tcl_FreeFile(errorId);
		}
		return NULL;
	    }
	}
    } else {

	/*
	 * Since we were called without a mode string, we need to compute
	 * the permissions according to whether an output pipe was
	 * created.
	 */

	permissions = ((outPipe != NULL) ? TCL_READABLE : 0);
    }
    
    pipePtr = (PipeState *) ckalloc((unsigned) sizeof(PipeState));

    pipePtr->inPipe = inPipe;		/* Output to pipe. */
    pipePtr->outPipe = outPipe;		/* Input from pipe. */
    pipePtr->errorId = errorId;		/* Error output from pipe. */
    pipePtr->numPids = numPids;		/* How many PIDs are
                                         * attached to the pipe? */
    pipePtr->pidPtr = pidPtr;		/* The PIDs themselves. */

    if (permissions & TCL_WRITABLE) {
        pipeId = inPipe;
    } else if (permissions & TCL_READABLE) {
        pipeId = outPipe;
    } else {
        pipeId = errorId;
    }
    if (pipeId == (Tcl_File) NULL) {
        channelId = 0;
    } else {
        channelId = (int) Tcl_GetFileInfo(pipeId, NULL);
    }
        
    /*
     * For backward compatibility with previous versions of Tcl, we
     * use "file%d" as the base name for pipes even though it would
     * be more natural to use "pipe%d".
     */

    sprintf(channelName, "file%d", channelId);
    chan = Tcl_CreateChannel(&pipeChannelType, channelName,
            (permissions & TCL_READABLE) ? outPipe : NULL,
            (permissions & TCL_WRITABLE) ? inPipe : NULL,
            (ClientData) pipePtr);
    if (chan == (Tcl_Channel) NULL) {
        Tcl_AppendResult(interp, "pipe for command could not be created",
                (char *) NULL);
        if (numPids > 0) {
            Tcl_DetachPids(numPids, pidPtr);
            ckfree((char *) pidPtr);
        }
        if (inPipe != NULL) {
            fd = (int) Tcl_GetFileInfo(inPipe, NULL);
            close(fd);
            Tcl_FreeFile(inPipe);
        }
        if (outPipe != NULL) {
            fd = (int) Tcl_GetFileInfo(outPipe, NULL);
            close(fd);
            Tcl_FreeFile(outPipe);
        }
        if (errorId != NULL) {
            fd = (int) Tcl_GetFileInfo(errorId, NULL);
            close(fd);
            Tcl_FreeFile(errorId);
        }
        return NULL;
    }
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_PidCmd --
 *
 *	This procedure is invoked to process the "pid" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_PidCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to get pids for. */
    Tcl_ChannelType *chanTypePtr;	/* The type of that channel. */
    PipeState *pipePtr;			/* The pipe state. */
    int i;				/* Loops over PIDs attached to the
                                         * pipe. */
    char string[50];			/* Temp buffer for string rep. of
                                         * PIDs attached to the pipe. */

    if (argc > 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " ?channelId?\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (argc == 1) {
	sprintf(interp->result, "%ld", (long) getpid());
    } else {
        chan = Tcl_GetChannel(interp, argv[1], NULL);
        if (chan == (Tcl_Channel) NULL) {
	    return TCL_ERROR;
	}
	chanTypePtr = Tcl_GetChannelType(chan);
	if (chanTypePtr != &pipeChannelType) {
	    return TCL_OK;
	}
        pipePtr = (PipeState *) Tcl_GetChannelInstanceData(chan);
        for (i = 0; i < pipePtr->numPids; i++) {
	    sprintf(string, "%d", pipePtr->pidPtr[i]);
	    Tcl_AppendElement(interp, string);
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateSocket --
 *
 *	This function opens a new socket in client or server mode
 *	and initializes the TcpState structure.
 *
 * Results:
 *	Returns a new TcpState, or NULL with an error interp.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static TcpState *
CreateSocket(interp, port, host, server)
    Tcl_Interp *interp;		/* For error reporting; can be NULL. */
    int port;			/* Port number to open. */
    char *host;			/* Name of host on which to open port. */
    int server;			/* 1 if socket should be a server socket,
				 * else 0 for a client socket. */
{
    int status, sock;
    struct sockaddr_in sockaddr;	/* Socket address */
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */
    struct linger ling;			/* For making sockets linger. */
    TcpState *statePtr;

    (void) memset((char *) &sockaddr, '\0', sizeof(struct sockaddr_in));
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_port = htons((unsigned short) (port & 0xFFFF));
    hostent = gethostbyname(host);
    if (hostent != NULL) {
	memcpy((char *) &addr,
	    (char *) hostent->h_addr_list[0], (size_t) hostent->h_length);
    } else {
	addr.s_addr = inet_addr(host);
	if (addr.s_addr == (unsigned long)-1) {
	    goto error;
	}
    }

    /*
     * There is a rumor that this assignment may require care on
     * some 64 bit machines.
     */

    sockaddr.sin_addr.s_addr = addr.s_addr;


    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	goto error;
    }

    /*
     * Set kernel space buffering
     */

    TclSockMinimumBuffers(sock, CHANNEL_BUF_SIZE);

    /*
     * Make socket close down gracefully.
     */
    
    ling.l_onoff = 1;
    ling.l_linger = 0;
    (void) setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &ling,
            sizeof(ling));

    if (server) {
	/*
	 * Set up to reuse server addresses automatically and bind to the
	 * specified port.
	 */
    
	status = 1;
	(void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
		sizeof(status));
	status = bind(sock, (struct sockaddr *) &sockaddr,
                sizeof(struct sockaddr));
	if (status != -1) {
	    status = listen(sock, 5);
	} 
    } else {
	/*
	 * Attempt to connect. The connect may fail at present with an
	 * EINPROGRESS but at a later time it will complete. The caller
	 * will set up a file handler on the socket if she is interested in
	 * being informed when the connect completes.
	 */

	status = connect(sock, (struct sockaddr *) &sockaddr,
		sizeof(sockaddr));
    }

    if (status < 0) {
        if (interp != NULL) {
            Tcl_AppendResult(interp, "couldn't open socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        close(sock);
        return NULL;
    }

    /*
     * Allocate a new TcpState for this socket.
     */

    statePtr = (TcpState *) ckalloc((unsigned) sizeof(TcpState));
    statePtr->sock = Tcl_GetFile((ClientData) sock, TCL_UNIX_FD);

    return statePtr;

  error:
    if (interp != NULL) {
	Tcl_AppendResult(interp, "couldn't open socket: ",
		Tcl_PosixError(interp), (char *) NULL);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpClient --
 *
 *	Opens a TCP client socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.  An error message is returned
 *	in the interpreter on failure.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpClient(interp, port, host)
    Tcl_Interp *interp;			/* For error reporting; can be NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Host on which to open port. */
{
    Tcl_Channel chan;
    TcpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateSocket(interp, port, host, 0);
    if (statePtr == NULL) {
	return NULL;
    }

    statePtr->acceptProc = NULL;
    statePtr->acceptProcData = (ClientData) NULL;

    sprintf(channelName, "sock%d",
	    (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, statePtr->sock,
	    statePtr->sock, (ClientData) statePtr);

    /*
     * Sockets start out with EOL translation set to "auto".
     */

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenTcpServer --
 *
 *	Opens a TCP server socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.
 *
 * Side effects:
 *	Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenTcpServer(interp, port, host, acceptProc, acceptProcData)
    Tcl_Interp *interp;			/* For error reporting - may be
                                         * NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Name of local host. */
    Tcl_TcpAcceptProc *acceptProc;	/* Callback for accepting connections
                                         * from new clients. */
    ClientData acceptProcData;		/* Data for the callback. */
{
    Tcl_Channel chan;
    TcpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateSocket(interp, port, host, 1);
    if (statePtr == NULL) {
	return NULL;
    }

    statePtr->acceptProc = acceptProc;
    statePtr->acceptProcData = acceptProcData;

    /*
     * Set up the callback mechanism for accepting connections
     * from new clients. The caller will use Tcl_TcpRegisterCallback
     * to register a callback to call when a new connection is
     * accepted.
     */

    Tcl_CreateFileHandler(statePtr->sock, TCL_READABLE, TcpAccept,
            (ClientData) statePtr);
    
    sprintf(channelName, "sock%d",
	    (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    chan = Tcl_CreateChannel(&tcpChannelType, channelName, NULL, NULL,
            (ClientData) statePtr);

    /*
     * Sockets start out with EOL translation set to "auto".
     */
    
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * TcpAccept --
 *	Accept a TCP socket connection.  This is called by the event loop,
 *	and it in turns calls any registered callbacks for this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Evals the Tcl script associated with the server socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
TcpAccept(data, mask)
    ClientData data;			/* Callback token. */
    int mask;				/* Not used. */
{
    TcpState *sockState;		/* Client data of server socket. */
    int newsock;			/* The new client socket */
    Tcl_File newFile;			/* Its file. */
    TcpState *newSockState;		/* State for new socket. */
    struct sockaddr_in addr;		/* The remote address */
    int len;				/* For accept interface */
    Tcl_Channel chan;			/* Channel instance created. */
    char channelName[20];

    sockState = (TcpState *) data;

    len = sizeof(struct sockaddr_in);
    newsock = accept((int) Tcl_GetFileInfo(sockState->sock, NULL),
	    (struct sockaddr *)&addr, &len);
    if (newsock < 0) {
        return;
    }
    
    newFile = Tcl_GetFile((ClientData) newsock, TCL_UNIX_FD);
    if (newFile) {
        newSockState = (TcpState *) ckalloc((unsigned) sizeof(TcpState));

        newSockState->sock = newFile;
        newSockState->acceptProc = (Tcl_TcpAcceptProc *) NULL;
        newSockState->acceptProcData = (ClientData) NULL;
        
        sprintf(channelName, "sock%d", (int) newsock);
        chan = Tcl_CreateChannel(&tcpChannelType, channelName, newFile,
                newFile, (ClientData) newSockState);
        if (chan == (Tcl_Channel) NULL) {
            ckfree((char *) newSockState);
            close(newsock);
            Tcl_FreeFile(newFile);
	} else {

            /*
             * Sockets start out in EOL translation mode "auto".
             */            

            if (sockState->acceptProc != (Tcl_TcpAcceptProc *) NULL) {
                (sockState->acceptProc) (sockState->acceptProcData, chan,
                        inet_ntoa(addr.sin_addr));
            }
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetStdChannels --
 *
 *	Returns handles to the standard channels for stdin, stdout and
 *	stderr.
 *
 * Results:
 *	None; returns values through output parameters.
 *
 * Side effects:
 *	May cause the creation of these standard channels and the underlying
 *	files.
 *
 *----------------------------------------------------------------------
 */

void
TclGetStdChannels(stdinPtr, stdoutPtr, stderrPtr)
    Tcl_Channel *stdinPtr,		/* Where to store the channels. */
    	        *stdoutPtr,
    		*stderrPtr;
{
    /*
     * If the channels were not created yet, create them now and
     * store them in the static variables.
     */

    if (stdinPtr != NULL) {
	if (stdinFile == NULL) {
	    stdinFile = Tcl_GetFile((ClientData) 0, TCL_UNIX_FD);
	    stdinChannel = CreateStdFileChannel("stdin", stdinFile, NULL);
        }
        *stdinPtr = (Tcl_Channel) stdinChannel;
    }
    if (stdoutPtr != NULL) {
	if (stdoutFile == NULL) {
	    stdoutFile = Tcl_GetFile((ClientData) 1, TCL_UNIX_FD);
	    stdoutChannel = CreateStdFileChannel("stdout", NULL, stdoutFile);
        }
	*stdoutPtr = (Tcl_Channel) stdoutChannel;
    }
    if (stderrPtr != NULL) {
        if (stderrFile == NULL) {
	    stderrFile = Tcl_GetFile((ClientData) 2, TCL_UNIX_FD);
	    stderrChannel = CreateStdFileChannel("stderr", NULL, stderrFile);
        }
        *stderrPtr = (Tcl_Channel) stderrChannel;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CreateStdFileChannel --
 *
 *	Create a channel suitable for use in wrapping a standard
 *	file handle.  Sets up the channel for use by the file driver.
 *
 * Results:
 *	Returns a newly allocated file channel.
 *
 * Side effects:
 *	Registers a new channel with the given name.
 *
 *----------------------------------------------------------------------
 */

static Tcl_Channel
CreateStdFileChannel(name, inFile, outFile)
    char *name;			/* Name of standard channel. */
    Tcl_File inFile;		/* File to use for input, or NULL. */
    Tcl_File outFile;		/* File to use for output, or NULL. */
{
    Tcl_Channel channel;

    channel = Tcl_CreateChannel(&fileChannelType, name, inFile, outFile,
	    (ClientData) NULL);

    /*
     * Set up the normal channel options for stdio handles.
     */

    if (channel != NULL) {
	Tcl_SetChannelOption(NULL, channel, "-translation", "lf");
	Tcl_SetChannelOption(NULL, channel, "-linemode", "on");
    }
    return channel;
}
