/* 
 * tclWinChan.c
 *
 *	Channel drivers for Windows channels based on files, command
 *	pipes and TCP sockets.
 *
 * Copyright (c) 1995-1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclWinChan.c 1.21 96/01/24 18:58:14
 */

#include	"tclInt.h"	/* Internal definitions for Tcl. */
#include	"tclPort.h"	/* Portability features for Tcl. */

/*
 * Static routines for this file:
 */

static int		CommonBufferSize _ANSI_ARGS_((
    			    ClientData instanceData));
static Tcl_Channel	CreateStdFileChannel _ANSI_ARGS_((char *name,
			    Tcl_File inFile, Tcl_File outFile));
static int		FilePipeBlockMode _ANSI_ARGS_((
    			    ClientData instanceData, Tcl_File inFile,
			    Tcl_File outFile, int mode));
static int		FileClose _ANSI_ARGS_((ClientData instanceData,
	            	    Tcl_File inFile, Tcl_File outFile));
static int		FileSeek _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File inFile, Tcl_File outFile, long offset,
			    int mode, int *errorCode));
static int		FilePipeInput _ANSI_ARGS_((ClientData instanceData,
	            	    Tcl_File inFile, char *buf, int toRead,
	            	    int *errorCode));
static int		FilePipeOutput _ANSI_ARGS_((ClientData instanceData,
			    Tcl_File outFile, char *buf, int toWrite,
			    int *errorCode));
static Tcl_File		GetStdFile _ANSI_ARGS_((DWORD handleId));
static int		PipeClose _ANSI_ARGS_((ClientData instanceData,
	            	    Tcl_File inFile, Tcl_File outFile));

/*
 * Static variables to hold channels for stdin, stdout and stderr.
 */

static Tcl_Channel stdinChannel = NULL;
static Tcl_Channel stdoutChannel = NULL;
static Tcl_Channel stderrChannel = NULL;

/*
 * Static variables to hold files for stdin, stdout and stderr.
 */

static Tcl_File stdinFile = NULL;
static Tcl_File stdoutFile = NULL;
static Tcl_File stderrFile = NULL;

/*
 * This structure describes the channel type structure for file based IO.
 */

static Tcl_ChannelType fileChannelType = {
    "file",			/* Type name. */
    FilePipeBlockMode,		/* Set blocking or non-blocking mode.*/
    CommonBufferSize,		/* Buffer size. */
    FileClose,			/* Close proc. */
    FilePipeInput,		/* Input proc. */
    FilePipeOutput,		/* Output proc. */
    FileSeek			/* Seek proc. */
};

/*
 * This structure describes the channel type structure for command pipe
 * based IO.
 */

static Tcl_ChannelType pipeChannelType = {
    "pipe",			/* Type name. */
    FilePipeBlockMode,		/* Set blocking or non-blocking mode.*/
    CommonBufferSize,		/* Buffer size. */
    PipeClose,			/* Close proc. */
    FilePipeInput,		/* Input proc. */
    FilePipeOutput,		/* Output proc. */
    NULL			/* Seek proc. */
};

/*
 * This is the size of the channel name for File based channels
 */

#define CHANNEL_NAME_SIZE	64
static char channelName[CHANNEL_NAME_SIZE+1];

/*
 * The preferred buffer size for Windows channels.
 */

#define CHANNEL_BUF_SIZE	4096

/*
 * Structure describing per-instance state for file based channels.
 *
 * IMPORTANT NOTE: If you modify this structure, make sure that the
 * "asynch" field remains the first field - FilePipeBlockMode depends
 * on this.
 */

typedef struct FileState {
    int asynch;			/* 1 if channel is in asynch mode. */
    int append;			/* 1 if channel is in append mode. */
} FileState;

/*
 * This structure describes per-instance state of a pipe based channel.
 *
 * IMPORTANT NOTE: If you modify this structure, make sure that the
 * "asynch" field remains the first field - FilePipeBlockMode depends
 * on this.
 */

typedef struct PipeState {
    int asynch;			/* 1 if channel is in asynch mode. */
    Tcl_File inPipe;		/* Output from pipe. */
    Tcl_File outPipe;		/* Input from pipe. */
    Tcl_File errorId;		/* Error output from pipe. */
    int numPids;		/* Number of processes attached to pipe. */
    int *pidPtr;		/* Pids of attached processes. */
} PipeState;


/*
 *----------------------------------------------------------------------
 *
 * CommonBufferSize --
 *
 *	Returns the preferred buffer size for IO channels.
 *
 * Results:
 *	The preferred buffer size.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
CommonBufferSize(clientData)
    ClientData clientData;		/* Unused. */
{
    return CHANNEL_BUF_SIZE;
}

/*
 *----------------------------------------------------------------------
 *
 * FilePipeOutput--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
FilePipeOutput(instanceData, outFile, buf, toWrite, errorCode)
    ClientData instanceData;		/* Unused. */
    Tcl_File outFile;			/* Output device for channel. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    int type;
    DWORD bytesWritten;
    HANDLE handle;
    
    *errorCode = 0;
    handle = (HANDLE) Tcl_GetFileInfo(outFile, &type);

    /*
     * If we are writing to a file that was opened with O_APPEND, we need to
     * seek to the end of the file before writing the current buffer.
     */

    if (type == TCL_WIN_FILE) {
	FileState *statePtr = (FileState *)instanceData;
	if (statePtr->append) {
	    SetFilePointer(handle, 0, NULL, FILE_END);
	}
    }

    if (WriteFile(handle, (LPVOID) buf, (DWORD) toWrite, &bytesWritten,
            (LPOVERLAPPED) NULL) == FALSE) {
        TclWinConvertError(GetLastError());
        if (errno == EPIPE) {
            return 0;
        }
        *errorCode = errno;
        return -1;
    }
    if (type == TCL_WIN_FILE) {
	FlushFileBuffers(handle);
    }
    return bytesWritten;
}

/*
 *----------------------------------------------------------------------
 *
 * FilePipeInput --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

static int
FilePipeInput(instanceData, inFile, buf, bufSize, errorCode)
    ClientData instanceData;		/* File state. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    FileState *statePtr;
    HANDLE handle;
    DWORD count;
    DWORD bytesRead;
    int type;

    *errorCode = 0;
    statePtr = (FileState *) instanceData;
    handle = (HANDLE) Tcl_GetFileInfo(inFile, &type);

    if (type == TCL_WIN_PIPE) {

	/*
	 * Pipes will block until the requested number of bytes has been
	 * read.  To avoid blocking unnecessarily, we look ahead and only
	 * read as much as is available.
	 */

	if (PeekNamedPipe(handle, (LPVOID) NULL, (DWORD) 0, (LPDWORD) NULL,
		&count, (LPDWORD) NULL) == TRUE) {
	    if ((count != 0) && ((DWORD) bufSize > count)) {
		bufSize = (int) count;
	    } else if ((count == 0) && statePtr->asynch) {
		errno = *errorCode = EAGAIN;
		return 0;
	    } else if ((count == 0) && !statePtr->asynch) {
		bufSize = 1;
	    }
	} else {
	    goto error;
	}
    }

    /*
     * Note that we will block on reads from a console buffer until a
     * full line has been entered.  The only way I know of to get
     * around this is to write a console driver.  We should probably
     * do this at some point, but for now, we just block.
     */

    if (ReadFile(handle, (LPVOID) buf, (DWORD) bufSize, &bytesRead,
            (LPOVERLAPPED) NULL) == FALSE) {
	goto error;
    }
    
    return bytesRead;

    error:
    TclWinConvertError(GetLastError());
    if (errno == EPIPE) {
	return 0;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * FilePipeBlockMode --
 *
 *	Set blocking or non-blocking mode on channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

static int
FilePipeBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Instance state for channel. */
    Tcl_File inFile, outFile;		/* Input, output for channel. */
    int mode;				/* The mode to set. */
{
    /*
     * Files on Windows can not be switched between blocking and nonblocking,
     * hence we have to emulate the behavior. This is done in the input
     * function by checking against a bit in the state. We set or unset the
     * bit here to cause the input function to emulate the correct behavior.
     *
     * IMPORTANT NOTE:
     *
     * The use of the "asynch" field below relies on the assumption that it
     * will be located at the same offset (0) in the instanceData associated
     * with all types of channels using this routine.
     */

    if (instanceData != (ClientData) NULL) {
        FileState *sPtr = (FileState *) instanceData;
	sPtr->asynch = (mode == TCL_MODE_BLOCKING) ? 0 : 1;
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * FileClose --
 *
 *	Closes the IO channel.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the physical channel
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
FileClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* Pointer to FileState structure. */
    Tcl_File inFile;		/* Input side. */
    Tcl_File outFile;		/* Output side. */
{
    HANDLE handle;
    int type, errorCode = 0;

    /*
     * Invalidate the stdio cache if necessary.  Note that we assume that
     * the stdio file and channel pointers will become invalid at the same
     * time.
     */

    if (inFile == stdinFile) {
	stdinFile = NULL;
	stdinChannel = NULL;
    }
    if (outFile == stdoutFile) {
	stdoutFile = NULL;
	stdoutChannel = NULL;
    } else if (outFile == stderrFile) {
	stderrFile = NULL;
	stderrChannel = NULL;
    }

    if (instanceData != (ClientData) NULL) {
        ckfree((char *) instanceData);
    }

    if (inFile != NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(inFile, &type);

	/*
	 * Check for read/write file so we only close it once.
	 */

	if (inFile == outFile) {
	    outFile = NULL;
	}
        Tcl_FreeFile(inFile);

	if (CloseHandle(handle) == FALSE) {
	    TclWinConvertError(GetLastError());
	    errorCode = errno;
	}

    }
    if (outFile != NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(outFile, &type);
	Tcl_FreeFile(outFile);

	if (CloseHandle(handle) == FALSE) {
	    TclWinConvertError(GetLastError());
	    if (errorCode == 0) {
		errorCode = errno;
	    }
	}
    }
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * FileSeek --
 *
 *	Seeks on a file-based channel. Returns the new position.
 *
 * Results:
 *	-1 if failed, the new position if successful. If failed, it
 *	also sets *errorCodePtr to the error code.
 *
 * Side effects:
 *	Moves the location at which the channel will be accessed in
 *	future operations.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
FileSeek(instanceData, inFile, outFile, offset, mode, errorCodePtr)
    ClientData instanceData;			/* Unused. */
    Tcl_File inFile, outFile;			/* Input and output
                                                 * devices for channel. */
    long offset;				/* Offset to seek to. */
    int mode;					/* Relative to where
                                                 * should we seek? */
    int *errorCodePtr;				/* To store error code. */
{
    DWORD moveMethod;
    DWORD newPos;
    HANDLE handle;
    int type;

    *errorCodePtr = 0;
    if (inFile != (Tcl_File) NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(inFile, &type);
    } else if (outFile != (Tcl_File) NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(outFile, &type);
    } else {
        *errorCodePtr = EFAULT;
        return -1;
    }
    
    if (mode == SEEK_SET) {
        moveMethod = FILE_BEGIN;
    } else if (mode == SEEK_CUR) {
        moveMethod = FILE_CURRENT;
    } else {
        moveMethod = FILE_END;
    }

    newPos = SetFilePointer(handle, offset, NULL, moveMethod);
    if (newPos == 0xFFFFFFFF) {
        TclWinConvertError(GetLastError());
        return -1;
    }
    return newPos;
}

/*
 *----------------------------------------------------------------------
 *
 * TclCleanupCommandChannel --
 *
 *	Perform special cleanup for command channels.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May wait for child processes.
 *
 *----------------------------------------------------------------------
 */

int
TclCleanupCommandChannel(interp, chan, keepNewline)
    Tcl_Interp *interp;		/* Interpreter to report errors in. */
    Tcl_Channel chan;		/* Channel to clean up after. */
    int keepNewline;		/* Non-zero means keep the last newline
                                 * in interp->result. */
{
    int result;
    PipeState *pipePtr;
    Tcl_ChannelType *chanTypePtr;
    Tcl_Channel errChan;
    HANDLE handle;

    /*
     * Punt if the channel is not a command channel.
     */

    chanTypePtr = Tcl_GetChannelType(chan);
    if (chanTypePtr != &pipeChannelType) {
        return TCL_OK;
    }

    pipePtr = (PipeState *) Tcl_GetChannelInstanceData(chan);
    if (pipePtr->inPipe != NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(pipePtr->inPipe, NULL);

	if (CloseHandle(handle) == FALSE) {
	    TclWinConvertError(GetLastError());
            Tcl_FreeFile(pipePtr->inPipe);
	    return -1;
	}

        Tcl_FreeFile(pipePtr->inPipe);
        pipePtr->inPipe = (Tcl_File) NULL;
    }
    if (pipePtr->outPipe != NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(pipePtr->outPipe, NULL);

	if (CloseHandle(handle) == FALSE) {
	    TclWinConvertError(GetLastError());
            Tcl_FreeFile(pipePtr->outPipe);
	    return -1;
	}

        Tcl_FreeFile(pipePtr->outPipe);
        pipePtr->outPipe = (Tcl_File) NULL;
    }

    /*
     * Wrap the error file into a channel and give it to the cleanup
     * routine.
     */

    if (pipePtr->errorId != NULL) {
	errChan = Tcl_CreateChannel(&fileChannelType, NULL, pipePtr->errorId,
		NULL, NULL);
    } else {
	errChan = NULL;
    }
    result = TclCleanupChildren(interp, pipePtr->numPids, pipePtr->pidPtr,
            errChan, keepNewline);
    
    pipePtr->numPids = 0;
    pipePtr->errorId = (Tcl_File) NULL;
    
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * PipeClose --
 *
 *	Closes a pipe based IO channel.
 *
 * Results:
 *	0 on success, errno otherwise.
 *
 * Side effects:
 *	Closes the physical channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
PipeClose(instanceData, inFile, outFile)
    ClientData instanceData;	/* Pointer to PipeState structure. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    PipeState *pipePtr = (PipeState *) instanceData;
    HANDLE handle;
    int i, errorCode = 0;

    if (pipePtr->inPipe != NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(pipePtr->inPipe, NULL);
        Tcl_FreeFile(pipePtr->inPipe);

	if (CloseHandle(handle) == FALSE) {
	    TclWinConvertError(GetLastError());
	    errorCode = errno;
	}
    }
    if (pipePtr->outPipe != NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(pipePtr->outPipe, NULL);
	Tcl_FreeFile(pipePtr->outPipe);

	if (CloseHandle(handle) == FALSE) {
	    TclWinConvertError(GetLastError());
	    if (errorCode == 0) {
		errorCode = errno;
	    }
	}
    }
    if (pipePtr->errorId != NULL) {
        handle = (HANDLE) Tcl_GetFileInfo(pipePtr->errorId, NULL);
	Tcl_FreeFile(pipePtr->errorId);

	if (CloseHandle(handle) == FALSE) {
	    TclWinConvertError(GetLastError());
	    if (errorCode == 0) {
		errorCode = errno;
	    }
	}
    }

    if (pipePtr->numPids > 0) {
        for (i = 0; i < pipePtr->numPids; i++) {
            if (pipePtr->pidPtr[i] != -1) {
                Tcl_DetachPids(1, &pipePtr->pidPtr[i]);
            }
        }
        ckfree((char *) pipePtr->pidPtr);
    }
    ckfree((char *) pipePtr);
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenFileChannel --
 *
 *	Open an File based channel on Unix systems.
 *
 * Results:
 *	The new channel or NULL. If NULL, the output argument
 *	errorCodePtr is set to a POSIX error.
 *
 * Side effects:
 *	May open the channel and may cause creation of a file on the
 *	file system.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenFileChannel(interp, fileName, modeString, permissions)
    Tcl_Interp *interp;			/* Interpreter for error reporting;
                                         * can be NULL. */
    char *fileName;			/* Name of file to open. */
    char *modeString;			/* A list of POSIX open modes or
                                         * a string such as "rw". */
    int permissions;			/* If the open involves creating a
                                         * file, with what modes to create
                                         * it? */
{
    Tcl_File file;
    Tcl_Channel chan;
    FileState *sPtr;
    int seekFlag, mode;
    HANDLE handle;
    DWORD accessMode, createMode, shareMode, flags;
    SECURITY_ATTRIBUTES sec;
    char *nativeName;
    Tcl_DString buffer;

    mode = TclGetOpenMode(interp, modeString, &seekFlag);
    if (mode == -1) {
        return NULL;
    }
    switch (mode & (O_RDONLY | O_WRONLY | O_RDWR)) {
	case O_RDONLY:
	    accessMode = GENERIC_READ;
	    break;
	case O_WRONLY:
	    accessMode = GENERIC_WRITE;
	    break;
	case O_RDWR:
	    accessMode = (GENERIC_READ | GENERIC_WRITE);
	    break;
	default:
	    panic("Tcl_OpenFileChannel: invalid mode value");
	    break;
    }

    /*
     * Map the creation flags to the NT create mode.
     */

    switch (mode & (O_CREAT | O_EXCL | O_TRUNC)) {
	case (O_CREAT | O_EXCL):
	case (O_CREAT | O_EXCL | O_TRUNC):
	    createMode = CREATE_NEW;
	    break;
	case (O_CREAT | O_TRUNC):
	    createMode = CREATE_ALWAYS;
	    break;
	case O_CREAT:
	    createMode = OPEN_ALWAYS;
	    break;
	case O_TRUNC:
	case (O_TRUNC | O_EXCL):
	    createMode = TRUNCATE_EXISTING;
	    break;
	default:
	    createMode = OPEN_EXISTING;
	    break;
    }

    /*
     * If the file is being created, get the file attributes from the
     * permissions argument, else use the existing file attributes.
     */

    if (mode & O_CREAT) {
        if (permissions & S_IWRITE) {
            flags = FILE_ATTRIBUTE_NORMAL;
        } else {
            flags = FILE_ATTRIBUTE_READONLY;
        }
    } else {
	flags = GetFileAttributes(fileName);
        if (flags == 0xFFFFFFFF) {
	    flags = 0;
	}
    }

    /*
     * Set up the security attributes so this file is not inherited by
     * child processes.
     */

    sec.nLength = sizeof(sec);
    sec.lpSecurityDescriptor = NULL;
    sec.bInheritHandle = 0;

    /*
     * Set up the file sharing mode.  We want to allow simultaneous access.
     */

    shareMode = FILE_SHARE_READ | FILE_SHARE_WRITE;

    /*
     * Now we get to create the file.
     */

    nativeName = Tcl_TranslateFileName(interp, fileName, &buffer);
    if (nativeName == NULL) {
	return NULL;
    }
    handle = CreateFile(nativeName, accessMode, shareMode, &sec, createMode,
            flags, (HANDLE) NULL);
    Tcl_DStringFree(&buffer);

    if (handle == INVALID_HANDLE_VALUE) {
	DWORD err = GetLastError();
	if ((err & 0xffffL) == ERROR_OPEN_FAILED) {
	    err = (mode & O_CREAT) ? ERROR_FILE_EXISTS : ERROR_FILE_NOT_FOUND;
	}
        TclWinConvertError(err);
	if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "couldn't open \"", fileName, "\": ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        return NULL;
    }

    file = Tcl_GetFile((ClientData) handle, TCL_WIN_FILE);

    sPtr = (FileState *) ckalloc((unsigned) sizeof(FileState));
    sPtr->asynch = 0;
    sPtr->append = (mode & O_APPEND) ? 1 : 0;
    sprintf(channelName, "file%d", (int) Tcl_GetFileInfo(file, NULL));
    chan = Tcl_CreateChannel(&fileChannelType, channelName,
            (accessMode & GENERIC_READ) ? file : NULL,
            (accessMode & GENERIC_WRITE) ? file : NULL,
            (ClientData) sPtr);
    if (chan == (Tcl_Channel) NULL) {
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp, "could not open channel \"",
                    channelName, "\": ", Tcl_PosixError(interp),
                    (char *) NULL);
        }
        Tcl_FreeFile(file);
        CloseHandle(handle);
        ckfree((char *) sPtr);
        return NULL;
    }

    if (seekFlag) {
        if (Tcl_Seek(chan, 0, SEEK_END) < 0) {
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "could not seek to end of file on \"",
                        channelName, "\": ", Tcl_PosixError(interp),
                        (char *) NULL);
            }
            Tcl_Close(chan);
            return NULL;
        }
    }
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenCommandChannel --
 *
 *	Opens a communication channel to a sequence of commands. 
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Creates processes, opens pipes.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
Tcl_Channel
Tcl_OpenCommandChannel(interp, argc, argv, modeString)
    Tcl_Interp *interp;		/* Interpreter for error reporting. */
    int argc;			/* How many arguments. */
    char **argv;		/* Array of arguments for command pipe. */
    char *modeString;		/* Which modes to open the channel in? */
{
    int permissions, channelId, mode, numPids, *pidPtr, seekFlag;
    Tcl_File *inPipePtr, *outPipePtr, inPipe, outPipe, errorId, pipeId;
    PipeState *pipePtr;
    Tcl_Channel chan;
    HANDLE handle;

    inPipe = outPipe = errorId = NULL;

    /*
     * If the modeString is NULL, then we want either a readable or a null
     * channel.  This is used by exec; it can't tell what mode to request
     * because there may be redirections.  Otherwise we parse the mode
     * string and set the permissions appropriately.
     */

    if (modeString == NULL) {
	outPipePtr = &outPipe;
	inPipePtr = NULL;
    } else {
        mode = TclGetOpenMode(interp, modeString, &seekFlag);
        if (mode == -1) {
            return (Tcl_Channel) NULL;
        }
	switch (mode & (O_RDONLY | O_WRONLY | O_RDWR)) {
	    case O_RDONLY:
		permissions = TCL_READABLE;
		break;
	    case O_WRONLY:
		permissions = TCL_WRITABLE;
		break;
	    case O_RDWR:
		permissions = (TCL_READABLE | TCL_WRITABLE);
		break;
	    default:
		panic("Tcl_OpenCommandChannel: invalid mode value");
		break;
	}
	inPipePtr = (permissions & TCL_WRITABLE) ? &inPipe : NULL;
	outPipePtr = (permissions & TCL_READABLE) ? &outPipe : NULL;
    }
    
    numPids = Tcl_CreatePipeline(interp, argc, argv, &pidPtr, inPipePtr,
            outPipePtr, &errorId);

    if (numPids < 0) {
        if (inPipe != NULL) {
            handle = (HANDLE) Tcl_GetFileInfo(inPipe, NULL);
            CloseHandle(handle);
            Tcl_FreeFile(inPipe);
        }
        if (outPipe != NULL) {
            handle = (HANDLE) Tcl_GetFileInfo(outPipe, NULL);
            CloseHandle(handle);
            Tcl_FreeFile(outPipe);
        }
        if (errorId != NULL) {
            handle = (HANDLE) Tcl_GetFileInfo(errorId, NULL);
            CloseHandle(handle);
	    Tcl_FreeFile(errorId);
        }
        return NULL;
    }

    if (modeString != NULL) {
	
	/*
	 * Since we were called with a mode string, we need to
	 * verify that the permissions match the set of pipes
	 * that were created.
	 */

	if (permissions & TCL_READABLE) {
	    if (outPipe == NULL) {
		if (inPipe != NULL) {
		    handle = (HANDLE) Tcl_GetFileInfo(inPipe, NULL);
		    CloseHandle(handle);
		    Tcl_FreeFile(inPipe);
		}
		Tcl_AppendResult(interp, "can't read output from command:",
			" standard output was redirected", (char *) NULL);
		if (numPids > 0) {
		    Tcl_DetachPids(numPids, pidPtr);
		    ckfree((char *) pidPtr);
		}
		if (errorId != NULL) {
		    handle = (HANDLE) Tcl_GetFileInfo(errorId, NULL);
		    CloseHandle(handle);
		    Tcl_FreeFile(errorId);
		}
		return NULL;
	    }
	}
	if (permissions & TCL_WRITABLE) {
	    if (inPipe == NULL) {
		if (outPipe != NULL) {
		    handle = (HANDLE) Tcl_GetFileInfo(outPipe, NULL);
		    CloseHandle(handle);
		    Tcl_FreeFile(outPipe);
		}
		Tcl_AppendResult(interp, "can't write input to command:",
			" standard input was redirected", (char *) NULL);
		if (numPids > 0) {
		    Tcl_DetachPids(numPids, pidPtr);
		    ckfree((char *) pidPtr);
		}
		if (errorId != NULL) {
		    handle = (HANDLE) Tcl_GetFileInfo(errorId, NULL);
		    CloseHandle(handle);
		    Tcl_FreeFile(errorId);
		}
		return NULL;
	    }
	}
    } else {

	/*
	 * Since we were called without a mode string, we need to compute
	 * the permissions according to whether an output pipe was
	 * created.
	 */

	permissions = ((outPipe != NULL) ? TCL_READABLE : 0);
    }

    pipePtr = (PipeState *) ckalloc((unsigned) sizeof(PipeState));

    pipePtr->asynch = 0;
    pipePtr->inPipe = inPipe;		/* Output to pipe. */
    pipePtr->outPipe = outPipe;		/* Input from pipe. */
    pipePtr->errorId = errorId;		/* Error output from pipe. */
    pipePtr->numPids = numPids;		/* How many PIDs are
                                         * attached to the pipe? */
    pipePtr->pidPtr = pidPtr;		/* The PIDs themselves. */

    if (permissions & TCL_WRITABLE) {
        pipeId = inPipe;
    } else if (permissions & TCL_READABLE) {
        pipeId = outPipe;
    } else {
        pipeId = errorId;
    }
    if (pipeId == (Tcl_File) NULL) {
        channelId = 0;
    } else {
        channelId = (int) Tcl_GetFileInfo(pipeId, NULL);
    }
    
    /*
     * For backward compatibility with previous versions of Tcl, we
     * use "file%d" as the base name for pipes even though it would
     * be more natural to use "pipe%d".
     */

    sprintf(channelName, "file%d", channelId);
    chan = Tcl_CreateChannel(&pipeChannelType, channelName,
            (permissions & TCL_READABLE) ? outPipe : NULL,
            (permissions & TCL_WRITABLE) ? inPipe : NULL,
            (ClientData) pipePtr);
    if (chan == (Tcl_Channel) NULL) {
        Tcl_AppendResult(interp, "pipe for command could not be created",
                (char *) NULL);
        if (numPids > 0) {
            Tcl_DetachPids(numPids, pidPtr);
            ckfree((char *) pidPtr);
        }
        if (inPipe != NULL) {
            handle = (HANDLE) Tcl_GetFileInfo(inPipe, NULL);
            CloseHandle(handle);
            Tcl_FreeFile(inPipe);
        }
        if (outPipe != NULL) {
            handle = (HANDLE) Tcl_GetFileInfo(outPipe, NULL);
            CloseHandle(handle);
            Tcl_FreeFile(outPipe);
        }
        if (errorId != NULL) {
            handle = (HANDLE) Tcl_GetFileInfo(errorId, NULL);
            CloseHandle(handle);
            Tcl_FreeFile(errorId);
        }
        return NULL;
    }
    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_PidCmd --
 *
 *	This procedure is invoked to process the "pid" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_PidCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    Tcl_Channel chan;			/* The channel to get pids for. */
    Tcl_ChannelType *typePtr;
    PipeState *pipePtr;			/* The pipe state. */
    int i;				/* Loops over PIDs attached to the
                                         * pipe. */
    char string[50];			/* Temp buffer for string rep. of
                                         * PIDs attached to the pipe. */

    if (argc > 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " ?channelId?\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (argc == 1) {
	sprintf(interp->result, "%ld", (long) getpid());
    } else {
        chan = Tcl_GetChannel(interp, argv[1], NULL);
        if (chan == (Tcl_Channel) NULL) {
	    return TCL_ERROR;
	}
	typePtr = Tcl_GetChannelType(chan);
	if (typePtr != &pipeChannelType) {
            return TCL_OK;
        }
        pipePtr = (PipeState *) Tcl_GetChannelInstanceData(chan);
        for (i = 0; i < pipePtr->numPids; i++) {
	    sprintf(string, "%d", pipePtr->pidPtr[i]);
	    Tcl_AppendElement(interp, string);
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetStdChannels --
 *
 *	Returns handles to the standard channels for stdin, stdout and
 *	stderr.
 *
 * Results:
 *	None; returns values through output parameters.
 *
 * Side effects:
 *	May cause the creation of these standard channels and the underlying
 *	files.
 *
 *----------------------------------------------------------------------
 */

void
TclGetStdChannels(stdinPtr, stdoutPtr, stderrPtr)
    Tcl_Channel *stdinPtr,		/* Where to store the channels. */
    	        *stdoutPtr,
    		*stderrPtr;
{
    /*
     * If the channels were not created yet, create them now and
     * store them in the static variables.  Note that we create them as
     * TCL_WIN_PIPE, because they 
     */
            
    if (stdinPtr != (Tcl_Channel *) NULL) {
        if (stdinFile == NULL) {
	    stdinFile = GetStdFile(STD_INPUT_HANDLE);
	    stdinChannel = CreateStdFileChannel("stdin", stdinFile, NULL);
        }
        *stdinPtr = (Tcl_Channel) stdinChannel;
    }
    if (stdoutPtr != (Tcl_Channel *) NULL) {
        if (stdoutFile == NULL) {
	    stdoutFile = GetStdFile(STD_OUTPUT_HANDLE);
	    stdoutChannel = CreateStdFileChannel("stdout", NULL, stdoutFile);
        }
        *stdoutPtr = (Tcl_Channel) stdoutChannel;
    }
    if (stderrPtr != (Tcl_Channel *) NULL){
        if (stderrFile == NULL) {
	    stderrFile = GetStdFile(STD_ERROR_HANDLE);
	    stderrChannel = CreateStdFileChannel("stderr", NULL, stderrFile);
        }
        *stderrPtr = (Tcl_Channel) stderrChannel;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * GetStdFile --
 *
 *	This function retrieves one of the standard file handles and
 *	gets the Tcl_File assocaited with the handle.
 *
 * Results:
 *	Returns a new file handle of the proper type for the given std
 *	handle.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_File
GetStdFile(handleId)
    DWORD handleId;		/* Handle to retrieve. */
{
    int type;
    DWORD mode, htype;
    HANDLE handle = GetStdHandle(handleId);
    if (handle == INVALID_HANDLE_VALUE) {
	return NULL;
    }
    htype = GetFileType(handle);
    switch (htype) {
	case FILE_TYPE_CHAR:
	    type = TCL_WIN_CONSOLE;
	    break;
	case FILE_TYPE_DISK:
	    type = TCL_WIN_FILE;
	    break;
	case FILE_TYPE_PIPE:
	    type = TCL_WIN_PIPE;
	    break;
	default:
	    return NULL;
    }
    return Tcl_GetFile((ClientData) handle, type);
}

/*
 *----------------------------------------------------------------------
 *
 * CreateStdFileChannel --
 *
 *	Create a channel suitable for use in wrapping a standard
 *	file handle.  Sets up the channel for use by the file driver.
 *
 * Results:
 *	Returns a newly allocated file channel.
 *
 * Side effects:
 *	Registers a new channel with the given name.
 *
 *----------------------------------------------------------------------
 */

static Tcl_Channel
CreateStdFileChannel(name, inFile, outFile)
    char *name;			/* Name of standard channel. */
    Tcl_File inFile;		/* File to use for input, or NULL. */
    Tcl_File outFile;		/* File to use for output, or NULL. */
{
    Tcl_Channel channel;
    FileState *statePtr = (FileState *) ckalloc(sizeof(FileState));

    channel = Tcl_CreateChannel(&fileChannelType, name, inFile, outFile,
	    (ClientData) statePtr);
    if (channel == NULL) {
	ckfree((char *)statePtr);
	return NULL;
    }

    /*
     * Set up the normal channel options for stdio handles.
     */

    statePtr->asynch = 0;
    Tcl_SetChannelOption(NULL, channel, "-translation", "crlf");
    Tcl_SetChannelOption(NULL, channel, "-linemode", "on");
    return channel;
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetCommandPids --
 *
 *	Stores a list of the command PIDs for a command channel in
 *	interp->result.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Modifies interp->result.
 *
 *----------------------------------------------------------------------
 */

void
TclGetCommandPids(interp, chan)
    Tcl_Interp *interp;
    Tcl_Channel chan;
{
    PipeState *pipePtr;
    Tcl_ChannelType *chanTypePtr;
    int i;
    char buf[20];

    /*
     * Punt if the channel is not a command channel.
     */

    chanTypePtr = Tcl_GetChannelType(chan);
    if (chanTypePtr != &pipeChannelType) {
        return;
    }

    pipePtr = (PipeState *) Tcl_GetChannelInstanceData(chan);
    for (i = 0; i < pipePtr->numPids; i++) {
        sprintf(buf, "%d", pipePtr->pidPtr[i]);
        Tcl_AppendElement(interp, buf);
    }
}
